/* Copyright 1998 by Dexter Reid */

package com.nosuch.cooking;

import java.awt.*;
import java.awt.event.*;
import java.awt.Graphics.*;
import java.beans.*;
import java.util.*;
import com.nosuch.midi.*;

/**
 * This is view control that can be used to visualize and
 * manipulate MidiPhrase structures.
 * 
 * @author Dexter Reid
 */
public class NSPhraseView extends Panel implements MouseListener {

	MidiPhrase m_phrase;
	Color [] m_color;
	float [] m_zoom_x;
	float [] m_zoom_y;
	boolean m_b_showing;
	// NSPhraseViewFilter m_filter;

	//
	// Things to color:
	//  these values can all be passed to setColor()
	//

	public final int IDX_BKGND = 0;
	public final int IDX_PHRASE = 1;
	public final int IDX_SELECTION = 2;


	//
	// Arrays that keep the Min and max time and pitch values being
	//  displayed
	long [] m_time_dim;
	int [] m_pitch_dim;

	//
	// Arrays that keep the selection bounds
	//
	long [] m_sel_time;
	int [] m_sel_pitch;

	//
	// Indicies for the time and pitch arrays
	//
	final int IDX_MIN = 0;
	final int IDX_MAX = 1;

	/**
	 * Performs some preliminary initialization of the view
	 * and adds the listener objects
	 */

	public NSPhraseView() {
		init();
		setLayout(null);
		setBackground(m_color[IDX_BKGND]);
		addMouseListener(this);
	}


	/**
	 * This uses double buffering to draw the phrase to the view.
	 * @param graphics he specified Graphics window
	 */

	public void paint(Graphics grph) {
		Dimension size = getSize();
		Image dbl_buffer = createImage(size.width,size.height);
		Graphics dbl_g = dbl_buffer.getGraphics();

		// If the view was not visible at the time that the phrase was
		// set, setPhrase() gets called again to adjust the zoom and pan
		if (!m_b_showing) {
			setPhrase(m_phrase,
				true,	// the true is so that the zoom is set to fit the phrase
				false);	// the false is so that repaint() is NOT called
			m_b_showing = true;
		}

		//
		// Draws the phrase to the off screen buffer
		if (m_phrase!=null)
			drawPhrase(dbl_g);

		//
		// Transfers the drawn phrase from the buffer to the display
		grph.drawImage(dbl_buffer, 0, 0, this);
	}

	/**
	 * This sets the phrase that the view will render.
	 * If the view is showing, repaint() will be called.
	 */

	public void setPhrase(MidiPhrase phrase, boolean b_fit_zoom) {
		setPhrase(phrase, b_fit_zoom, true);
	}

	/**
	 * This gets the phrase that view is currently rendering
	 */

	public MidiPhrase getPhrase() {
		return m_phrase;
	}


	/*
     * This sets the filter to remove MidiMsgs from the view
    public void setViewFilter(NSPhraseViewFilter filter)
    {
        m_filter = filter;
    }
	/*
     * This gets the filter that is currently being used to remove MidiMsgs from the view
    public NSPhraseViewFilter getViewFilter()
    {
    }
	 */

	/**
	 * Sets the color of the given view attribute.
	 * 
	 * 
	 * @param attribute the attribute of the view whose color is to be set
	 * @param color shade to set the view attribute
	 */

	public void setColor(int attribute, Color color) {
		m_color[attribute] = color;
		repaint();
	}

	/**
	 * Gets the color of the given view attribute
	 * 
	 * @param attribute the attribute of the view whose color is to be retrieved
	 * @return color of the attribute requested
	 */

	public Color getColor(int attribute) {
		return m_color[attribute];
	}

	/**
	 * Sets the view's selection
	 * 
	 * 
	 * @param point1 first point in bounding box of selection
	 * @param point2 second point in bounding box of selection
	 */

	public void setSelection(Point point1, Point point2) {
		int x1 = point1.x < point2.x ? point1.x:point2.x;
		int y1 = point1.y < point2.y ? point1.y:point2.y;
		int x2 = point1.x > point2.x ? point1.x:point2.x;
		int y2 = point1.y > point2.y ? point1.y:point2.y;

		m_sel_time[IDX_MIN] = convertXCoord(x1);
		m_sel_time[IDX_MAX] = convertXCoord(x2);
		m_sel_pitch[IDX_MIN] = convertYCoord(y1);
		m_sel_pitch[IDX_MAX] = convertYCoord(y2);

		repaint(x1, y1, (x2-x1), (y2-y1));
	}

	/**
	 * Pans the view so that the given XY coord becomes the upper-left corner
	 * 
	 */
	public void panXY(int x_pos, int y_pos) {
		long time =-1;
		int pitch =-1;

		if (x_pos != -1)
			time = convertXCoord(x_pos);
		if (y_pos !=-1)
			pitch = convertYCoord(y_pos);

		panTimePitch(time, pitch);
	}

	/**
	 * Pans the view so that the given XY coord becomes the upper-left corner
	 * 
	 */
	public void panXYRel(int x_pos, int y_pos) {
		long time =0;
		int pitch =0;

		Dimension size = getSize();
		time = convertXCoord(x_pos) - m_time_dim[IDX_MIN];
		pitch = convertYCoord(size.height - y_pos) - m_pitch_dim[IDX_MIN];

		panTimePitchRel(time, pitch);
	}

	/**
	 * Pans the view so that the upper-left corner is the specified time and pitch.
	 * 
	 */
	public void panTimePitch(long time, int pitch) {
		if (time != -1) {
			long time_range = m_time_dim[IDX_MAX] - m_time_dim[IDX_MIN];
			m_time_dim[IDX_MIN] = time;
			m_time_dim[IDX_MAX] = time + time_range;
		}
		if (pitch !=-1) {
			int pitch_range = m_pitch_dim[IDX_MAX] - m_pitch_dim[IDX_MIN];
			m_pitch_dim[IDX_MIN] = pitch;
			m_pitch_dim[IDX_MAX] = pitch + pitch_range;
		}
		zoomBase();
	}

	/**
	 * Pans the view with relative Time-Pitch coords
	 * 
	 */
	public void panTimePitchRel(long time, int pitch) {
		m_time_dim[IDX_MIN] += time;
		m_time_dim[IDX_MAX] += time;

		m_pitch_dim[IDX_MIN] += pitch;
		m_pitch_dim[IDX_MAX] += pitch;

		zoomBase();
		//repaint();
	}

	/**
	 * Sets the zoom so that the whole phrase fits inside
	 * the view.
	 * 
	 */

	public void zoomToFitPhrase() {
		// Get the pitch range
		// and the time range from the phrase
		m_time_dim[IDX_MIN] = 0;
		m_time_dim[IDX_MAX] = m_phrase.getLength();
		m_pitch_dim[IDX_MIN] = 0;
		m_pitch_dim[IDX_MAX] = 127;
		zoomTimePitch(m_time_dim, m_pitch_dim);
	}

	/**
	 * Sets the zoom for the time and pitch.
	 * the units are in ticks and notes
	 * 
	 */

	public void zoomTimePitch(long [] time_bounds, int [] pitch_bounds) {
		// Get the time range and the pitch range

		if (time_bounds != null) {
			m_time_dim[IDX_MIN] = time_bounds[IDX_MIN];
			m_time_dim[IDX_MAX] = time_bounds[IDX_MAX];
		}

		if (pitch_bounds != null) {
			m_pitch_dim[IDX_MIN] = pitch_bounds[IDX_MIN];
			m_pitch_dim[IDX_MAX] = pitch_bounds[IDX_MAX];
		}

		// Don't proceed if the phrase is null
		if (m_phrase == null)
			return;

		zoomBase();	// common code for all zoom functions
	}


	/**
	 * Sets the zoom in the x and y directions based on the
	 * current pitch and time values for the points passed
	 * into the function.
	 */

	public void zoomXY(Point point1, Point point2) {
		// Don't proceed if the phrase is null
		if (m_phrase == null)
			return;

		// Get the pitch range
		// and the time range from the x_bound and y_bound

		if (point1.x == -1) {
			m_time_dim[IDX_MIN] = convertXCoord(point1.x);
			m_time_dim[IDX_MAX] = convertXCoord(point2.x);
		}

		if (point1.y == -1) {
			m_pitch_dim[IDX_MIN] = convertYCoord(point1.y);
			m_pitch_dim[IDX_MAX] = convertYCoord(point2.y);
		}

		zoomBase();	// common code for all zoom functions
	}

	/**
	 * Sets the zoom in the x and y directions based on the
	 * current pitch and time values for the points passed
	 * into the function.
	 * 
	 * @param point1 x,y coordinate to start zoom
	 * @param percent 0-100 that represents the amount to zoom
	 */

	public void zoomPercentXY(Point point1, float percent) {
		// Don't proceed if the phrase is null
		if (m_phrase == null)
			return;

		// Get the pitch range
		// and the time range from the x_bound and y_bound

		Dimension size = getSize();
		float magnify;
		magnify = (float)(100.0/percent);
		if (point1.x != -1) {
			m_time_dim[IDX_MIN] = convertXCoord(point1.x);
			m_time_dim[IDX_MAX] = (long)(convertXCoord((int)(size.width * magnify)) );
		}

		if (point1.y != -1) {
			m_pitch_dim[IDX_MAX] = convertYCoord(point1.y);
			m_pitch_dim[IDX_MIN] = (int)(convertYCoord((int)(size.height * magnify)) );
		}

		zoomBase();
	}

	/**
	 * Draws a note on the screen in relation to the current
	 * zoom and panning of the phrase. It isn't necessary
	 * to call this method directly in most cases. Only if
	 * you want to start drawing notes on the screen that are
	 * not in the phrase or if you want to change the appearance
	 * of some of the existing notes.
	 * 
	 * @param pitch value of note to play
	 * @param start_time time that the note begins to play
	 * @param duration length of time for which the note plays
	 * @param color shade of the note to display
	 */

	public void drawNote(Graphics grph,
		int pitch,
		long start_time,
		long duration,
		Color color) {
		// convert start_time and note_length into coordinates
		int [] x_dim = convertTime(start_time, duration);
		int [] y_dim = convertPitch(pitch);

		// check for clipping
		if ( (( pitch < m_pitch_dim[IDX_MIN]) || (pitch > m_pitch_dim[IDX_MAX]))
			|| ((start_time > m_time_dim[IDX_MAX]) || (start_time + duration < m_time_dim[IDX_MIN])) )
			;	// clipped
		else {
			// draw a solid rectangle with the specified settings
			grph.setColor(color);
			grph.fillRect(x_dim[0],y_dim[0],x_dim[1],y_dim[1]);
		}
	}

	/**
	 * Converts a x coordinate into a time
	 * 
	 * @param x_coord position to convert into a time
	 * @return a unit of time
	 */

	public long convertXCoord(int x_coord) {
		return (long) (((float)x_coord / m_zoom_x[1]) + m_zoom_x[0]);
	}

	/**
	 * Converts a y_coord into a pitch
	 * 
	 * @param y_coord y coordinate to convert into a pitch
	 * @return pitch value that corresponds to the y_coordinate passed in
	 */

	public int convertYCoord(int y_coord) {
		return (int) (m_zoom_y[0] - ((float)y_coord / m_zoom_y[1]));
	}

	/**
	 * Converts a time into an x coordinates
	 * 
	 * @param time time to convert into an x pixel coordinate
	 * @return x-coordinate that corresponds to the time parameter
	 */

	int convertTime(long time) {
		return (int) (m_zoom_x[1] * (time - m_zoom_x[0])) ;
	}

	/**
	 * Converts a pitch into a pair of y coordinates
	 * 
	 * @param pitch pitch to convert into a pait of y pixel coordinates
	 * @return y-coordinate pair that corresponds to the pitch parameter
	 */
	int [] convertPitch(int pitch) {
		int [] pitch_dim = new int[2];
		pitch_dim[0] = (int) (m_zoom_y[1] * (m_zoom_y[0] - pitch));
		pitch_dim[1] = (int) m_zoom_y[1];
		return pitch_dim;
	}

	/**
	 * 
	 */
	public void mouseClicked(MouseEvent event) {
	}

	/**
	 * 
	 */
	public void mouseEntered(MouseEvent event) {
	}

	/**
	 * 
	 */

	public void mouseExited(MouseEvent event) {
	}

	/**
	 * 
	 */

	public void mousePressed(MouseEvent event) {
	}

	/**
	 * 
	 */

	public void mouseReleased(MouseEvent event) {
	}

	/**
	 * Initializes all of the member variables and the colors
	 * of the view's attributes.
	 */

	void init() {
		Dimension size = new Dimension(100,100);
		m_color = new Color[3];
		m_zoom_x = new float[2];
		m_zoom_y = new float[2];
		m_time_dim = new long[2];
		m_pitch_dim = new int[2];
		m_phrase = null;
		m_b_showing = false;

		//
		// Assign initial color values
		//
		m_color[IDX_BKGND] = Color.white;
		m_color[IDX_PHRASE] = Color.blue;
		m_color[IDX_SELECTION] = Color.yellow;

		// Initialize pitch range and time range to
		// whatever...we don't have a phrase to show yet
		m_time_dim[IDX_MIN] = 0;
		m_time_dim[IDX_MAX] = 100;
		m_pitch_dim[IDX_MIN] = 0;
		m_pitch_dim[IDX_MAX] = 127;
	}


	/**
	 * This sets the phrase that the view will render.
	 * Currently, there is a problem -- if b_fit_zoom is
	 * true, then the b_paint param doesn't matter.
	 * 
	 * @param b_paint repaints the view if true
	 */

	void setPhrase(MidiPhrase phrase, boolean b_fit_zoom, boolean b_paint) {
		MidiPhrase old_phrase = m_phrase;
		m_phrase = phrase;

		// Don't proceed if the phrase is null
		if (m_phrase == null)
			return;

		// if old_phrase was null, re-adjust the zoom so that it
		//  shows all of the new phrase

		if (b_fit_zoom)
			zoomToFitPhrase();
		else if (b_paint)
			repaint();
	}

	/**
	 * Draws the phrase that is currently set in the view
	 * 
	 */

	void drawPhrase(Graphics grph) {
		//
		// Clean up view
		Dimension size = getSize();
		//grph.clearRect(0,0,size.width,size.height);

		//
		// Loop and draw notes
		int duration=0;
		int start_time;
		int pitch;
		MidiMsg midi_msg;
		for ( Enumeration enum=m_phrase.getEnumeration(); enum.hasMoreElements(); ) {
			midi_msg = (MidiMsg) enum.nextElement();
			drawNote(grph,
				midi_msg.getPitch(),
				midi_msg.getTime(),
				midi_msg.getDuration(),
				m_color[IDX_PHRASE]);
			// System.out.println(midi_msg);
		}
	}

	/**
	 * Based on the pitch and time ranges,
	 * calculate the zoom variables
	 * 
	 */

	void zoomBase() {
		if (isShowing()) {
			Dimension size = getSize();
			m_zoom_x[0] = m_time_dim[IDX_MIN];
			m_zoom_y[0] = m_pitch_dim[IDX_MAX];	// -1
			m_zoom_x[1] = (float)size.width/(float)(m_time_dim[IDX_MAX]-m_time_dim[IDX_MIN]);
			m_zoom_y[1] = (float)size.height/(float)(m_pitch_dim[IDX_MAX]-m_pitch_dim[IDX_MIN]);
			if (m_b_showing)
				repaint();
		}
	}

	/**
	 * Converts a start_time/duration pair into two x pixel coordinates
	 * 
	 * @param start_time begining time to convert
	 * @param duration start_time + duration = end_time to convert
	 * @return a pair
	 */

	int [] convertTime(long start_time,
		long duration) {
		int [] time_dim = new int[2];
		time_dim[0] = (int) convertTime(start_time);
		time_dim[1] = (int) convertTime((int)m_zoom_x[0] + duration);
		return time_dim;
	}


	//
	//
	//
	//
	//
	// Test Functions: these are just functions i use for testing
	//  purposes
	//
	//
	//
	//
	//
	//

	void drawFakePhrase() {
		Dimension size = getSize();
		Graphics grph = getGraphics();
		grph.clearRect(0,0,size.width,size.height);
		//
		// Loop and draw notes
		int mode = 1;
		int duration=0;
		int start_time;
		int pitch;
		for(start_time=0; start_time < m_time_dim[IDX_MAX]; start_time+=duration) {
			pitch = (int) (Math.random() * m_pitch_dim[1]);
			duration = (int) (Math.random() * (m_time_dim[IDX_MAX] - start_time));
			drawNote(grph, pitch, start_time, duration, m_color[IDX_PHRASE]);
		}
	}

	void drawScalePhrase() {
		Dimension size = getSize();
		Graphics grph = getGraphics();
		grph.clearRect(0,0,size.width,size.height);
		//
		// Loop and draw notes
		int mode = 1;
		int duration=0;
		int start_time;
		int pitch=m_pitch_dim[0];
		duration = 10;
		for(start_time=0; start_time < m_time_dim[IDX_MAX]; start_time+=duration) {
			drawNote(grph, pitch, start_time, duration, m_color[IDX_PHRASE]);
			pitch++;
		}
	}

	boolean isClipped(int pitch, int start_time, int duration) {
		boolean ret_value = false;
		if (( pitch < m_pitch_dim[0]) || (pitch > m_pitch_dim[1]))
			ret_value = true;
		else if ((start_time < m_time_dim[IDX_MAX]) || (start_time + duration < m_time_dim[IDX_MIN]))
			ret_value = true;

		return ret_value;
	}

}

