/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

import java.util.Random;

/**
 * MidiAlgorithm collects some algorithmic composition functions.
 *
 * @author Tim Thompson
 */

public class Alg {

	/**
	 * Flip the input phrase.
	 */
	public static MidiPhrase flip(MidiPhrase p) {

		MidiPhrase r = new MidiPhrase();
		int fp = 64;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			m = (MidiMsg) m.clone();
			int pitch = m.getPitch();
			if ( pitch >= 0 )
				m.setPitch(fp - pitch);
			r.merge(m);
		}
		return r;
	}

	/**
	 * Arpeggiate the input phrase.
	 */
	public static MidiPhrase arpeggio(MidiPhrase p) {

		MidiPhrase r = new MidiPhrase();
		long lastend = 0;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			m = (MidiMsg) m.clone();
			m.setTime(lastend);
			r.merge(m);
			lastend += m.getDuration();
		}
		r.setLength(lastend);
		return r;
	}

	/**
	 * Repeat a phrase n times.
	 */
	public static MidiPhrase repeat(MidiPhrase p, int n) {
		return repeat(p,n,0);
	}

	public static MidiPhrase repeat(MidiPhrase p, int n, int q) {

		if ( n == 0 )
			return new MidiPhrase();

		// We only have to clone p if we have to change its length.
		if ( q > 0 ) {
			// quantize length of phrase we're going to repeat
			p = (MidiPhrase)p.clone();
			p.setLength( numquant(p.getLength(),q) );
			if ( p.getLength() == 0 )
				p.setLength(q);
		}
		MidiPhrase r = new MidiPhrase();
		while ( n-- > 0 )
			r.append(p);
		return r;
	}

	public static MidiPhrase echo(MidiPhrase ph,int num,int rtime,double rfactor) {
		// defaults for rtime and rfactor were 1b and 0.90
		ph = (MidiPhrase)ph.clone();	// se we can mess with the values
		long origt = ph.getNth(1).getTime();
		ph.addTime(-origt);
		MidiPhrase r = (MidiPhrase)ph.clone();
		for ( int n=0; n<num; n++ ) {
			ph.addTime(rtime);
			ph.multiplyVelocity(rfactor);
			r.merge(ph);
		}
		// r.reOrder();
		r.addTime(origt);
		return(r);
	}

	/**
	 * Return a phrase containing only "normal" notes in the input phrase.
	 */
	public static MidiPhrase onlyNotes(MidiPhrase p) {

		MidiPhrase r = new MidiPhrase();
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( m instanceof MidiNote
				|| m instanceof MidiNoteOn
				|| m instanceof MidiNoteOff ) {
				r.merge(m);
			}
		}
		return r;
	}

	/**
	 * Return a phrase containing "non" notes (e.g. MidiBytes)
	 * in the input phrase.
	 */
	public static MidiPhrase nonNotes(MidiPhrase p) {

		MidiPhrase r = new MidiPhrase();
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( ! ( m instanceof MidiNoteable ) ) {
				r.merge(m);
			}
		}
		return r;
	}

	/** Return the lowest pitch in the specified phrase. */
	public static int lowest(MidiPhrase p) {
		int low = 128;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( m instanceof MidiNoteable ) {
				if ( m.getPitch() < low )
					low = m.getPitch();
			}
		}
		return low;
	}

	/** Return the lowest-pitched note in the specified phrase. */
	public static MidiMsg lowestnt(MidiPhrase p) {
		int low = 128;
		MidiMsg lownt = null;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( m instanceof MidiNoteable ) {
				if ( m.getPitch() < low ) {
					low = m.getPitch();
					lownt = m;
				}
			}
		}
		if ( lownt != null )
			lownt = (MidiMsg)lownt.clone(); 
		return lownt;
	}

	/** Return the highest pitch in the specified phrase. */
	public static int highest(MidiPhrase p) {
		int high = -1;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( m instanceof MidiNoteable ) {
				if ( m.getPitch() > high )
					high = m.getPitch();
			}
		}
		return high;
	}

	/** Return the highest pitch in the specified phrase. */
	public static MidiMsg highestnt(MidiPhrase p) {
		int high = -1;
		MidiMsg highnt = null;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( m instanceof MidiNoteable ) {
				if ( m.getPitch() > high ) {
					high = m.getPitch();
					highnt = m;
				}
			}
		}
		if ( highnt != null )
			highnt = (MidiMsg)highnt.clone(); 
		return highnt;
	}

	/**
	 * findroot - given a phrase, try to guess the root of what's being played.
	 * The algorithm was given to me by Christopher John Rolfe (rolfe@sfu.ca),
	 * who says it was culled from W.Russo, Jazz Composition and Orchestration,
	 * p.25, ex.7.
	 */

	private static int[] Strength;
	static {
		Strength = new int[12];
		Strength[0]=12;
		Strength[1]=8;
		Strength[2]=6;
		Strength[3]=4;
		Strength[4]=2;
		Strength[5]=1;
		Strength[6]=12;
		Strength[7]=0;
		Strength[8]=3;
		Strength[9]=5;
		Strength[10]=7;
		Strength[11]=9;
	}

	public static MidiMsg findRoot(MidiPhrase ph) {
		ph.setTime(0);	// so that notes are sorted by pitch

		ph.reOrder();	// this may not really be needed, if we someday
						// guarantee that phrases are always in order

		int maxv = 999;
		MidiMsg root = null;
		ph = onlyNotes(ph);
		int n = ph.size();
		for ( int n1=1; n1<=n; n1++ ) {
			for ( int n2=n1; n2<=n; n2++ ) {
				MidiMsg m1 = ph.getNth(n1);
				MidiMsg m2 = ph.getNth(n2);
				int dp = m2.getPitch()-m1.getPitch();
				while ( dp >= 12 )
					dp -= 12;
				int v = Strength[dp];
				if ( v < maxv ) {
					if ( (v%2)==0 )
						root = ph.getNth(n1);
					else
						root = ph.getNth(n2);
					maxv = v;
				}
			}
		}
		return root;
	}

	/** Figure out a root note every intrvl beats */
	public static MidiPhrase makeRootEvery(MidiPhrase melody,int intrvl) {
		int t1 = 0;
		int t2 = intrvl;
		melody = onlyNotes(melody);
		MidiPhrase r = new MidiPhrase();
		for ( ; ; t2 += intrvl ) {
			MidiPhrase ct = cutTime(melody,t1,t2);
			if ( ct.size() == 0 ) {
				if ( cutTime(melody,t2).size() == 0 )
					break;
				t1 = t2;
				continue;
			}
			int low = lowest(ct);
			MidiMsg rt = findRoot(ct);
			if ( rt == null ) {
				t1 = t2;
				continue;
			}
			while ( rt.getPitch() >= low && rt.getPitch() > 12 )
				rt.setPitch(rt.getPitch() - 12);
			// chrd = chord(rt,"major")
			rt.setDuration(intrvl);
			rt.setTime(t1);
			r.merge(rt);
			t1 = t2;
		}
		return(r);
	}
	
	public static MidiPhrase addRootEvery(MidiPhrase mel,int intrvl) {
		return( mel.merge( makeRootEvery(mel,intrvl) ) );
	}

	/** Return a specified time range of a phrase */
	public static MidiPhrase cutTime(MidiPhrase p, long tm1, long tm2) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		MidiPhrase r = new MidiPhrase();
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			long t = m.getTime();
			if ( t < tm1 )
				continue;
			if ( t >= tm2 )
				break;
			r.merge(m);
		}
		return r;
	}

	/** Return all notes after a specified time */
	public static MidiPhrase cutTime(MidiPhrase p, long tm1) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		MidiPhrase r = new MidiPhrase();
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			long t = m.getTime();
			if ( t >= tm1 ) {
				r = (MidiPhrase)p.cloneStartingFrom(m);
				break;
			}
		}
		return r;
	}

	/** For each section of phrase ph of width wid, repeat it num times */
	public static MidiPhrase stutter(MidiPhrase ph, int wid,int num) {
		int t = 0;
		MidiPhrase r = new MidiPhrase();
		for (;;) {
			MidiPhrase p = Alg.cutTime(ph,t,t+wid);
			// if no notes left, we're done
			if ( p.size()==0 && Alg.cutTime(ph,t+wid).size()==0 )
				break;
			p.addTime(-t);
			p.setLength(wid);
			for ( int n=0; n<num; n++ )
				r.append(p);
			t += wid;
		}
		return(r);
	}
	
	// public static int rand(int n) {
	// 	Random r = new Random();
	// 	int i = r.nextInt();
	// 	if ( i < 0 )
	// 		i = -i;
	// 	return i%n;
	// }

	public static MidiPhrase merge(MidiPhrase p1, MidiPhrase p2) {
		p1 = (MidiPhrase)p1.clone();
		p1.merge(p2);
		return p1;
	}

	public static MidiPhrase merge(MidiPhrase p1, MidiPhrase p2, MidiPhrase p3){
		p1 = (MidiPhrase)p1.clone();
		p1.merge(p2);
		p1.merge(p3);
		return p1;
	}

	public static MidiPhrase merge(MidiPhrase p1, MidiPhrase p2,
					 MidiPhrase p3, MidiPhrase p4){
		p1 = (MidiPhrase)p1.clone();
		p1.merge(p2);
		p1.merge(p3);
		p1.merge(p4);
		return p1;
	}

	/** For each section (of random width, within a small range
	 * of possible widths) of phrase ph, repeat it
	 */
	public static MidiPhrase stutterrand(MidiPhrase ph) {
		int t = 0;
		MidiPhrase r = new MidiPhrase();
		Random rand = new Random();
		for (;;) {
			int n = rand.nextInt()%4;
			if ( n < 0 )
				n = -n;
			int b = (int)MidiSystem.clicksPerBeat();
			int wid;
			switch (n) {
			case 0: wid=1*b/2; break;
			case 1: wid=1*b; break;
			case 2: wid=2*b; break;
			case 3: default: wid=4*b; break;
			}
			MidiPhrase p = Alg.cutTime(ph,t,t+wid);
			// if no notes left, we're done
			if ( p.size()==0 && Alg.cutTime(ph,t+wid).size()==0 )
				break;
			p.addTime(-t);
			p.setLength(wid);
			r.append(p);
			r.append(p);
			t += wid;
			r.setLength(2*t);
		}
		return(r);
	}

	/** Transpos the pitch of the notes in a phrase by a given amount. */
	public static MidiPhrase transpose(MidiPhrase p, int delta) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		MidiPhrase r = new MidiPhrase();
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			m = (MidiMsg)m.clone();
			m.setPitch(m.getPitch()+delta);
			r.merge(m);
		}
		return r;
	}

	/** Return the latest ending time of the notes in a phrase. */
	public static long latest(MidiPhrase p) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		long maxtime = -1;
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			long endof = m.getTime() + m.getDuration();
			if ( endof > maxtime )
				maxtime = endof;
		}
		return maxtime;
	}

	public static MidiPhrase transposeSeqInPlace(MidiPhrase p, MidiPhrase seq) {
		MidiPhrase r = new MidiPhrase();
		int base = seq.getNth(1).getPitch();
		int num = seq.size();
		long lng = latest(p);
		long t1 = 0;
		for ( int i=1; i<=num; i++ ) {
			long t2 = (int)(((double)(i)*lng)/num);
			MidiPhrase sect = cutTime(p,t1,t2);
			int t = seq.getNth(i).getPitch() - base;
			r.merge(transpose(sect,t));
			t1 = t2;
		}
		return r;
	}

	/** Quantizes the specified number using a particular quantize value.
	 * If you want to guarantee that the result is not less than the
	 * original (which is typical when you are quantizing a time value
	 * and want to guarantee that the result is some time in the future),
	 * use nextquant().
	 */
	public static long numquant(long v, long q) {
		long rem;
		if ( v < 0 ) {
			rem = -v % q;
			if ( (rem*2) > q )
				v -= (q-rem);
			else
				v += rem;
		}
		else {
			rem = v % q;
			if ( (rem*2) > q )
				v += (q-rem);
			else
				v -= rem;
		}
		return(v);
	}

	/** This function quantizes the specified number using the specified
	 * quantize value, and returns the result.  The result is guaranteed to
	 * be greater than the original number; if quantization would produce
	 * a smaller number, it is increased by the quantize value.  This is
	 * typically used when you want to quantize a time value, and want to
	 * guarantee that the result will refer to some time in the future,
	 * not in the past.
	 */
	public static long nextquant(long t, long q) {
		long rem;
		long tq;
		if ( q == 1 )
			return(t);
		tq = t;
		rem = tq % q;
		if ( (rem*2) > q )
			tq += (q-rem);
		else
			tq -= rem;
		if ( tq < t )
			tq += q;
		return(tq);
	}

	/** Quantize a phrase.  Each note's starting time will be
	 * quantized by the specified quant value.  If a limit
	 * is specified, notes that would need to be moved by an
	 * amount larger than this limit will not be adjusted at all.
	 */
	public static MidiPhrase quantize(MidiPhrase ph,int qnt) {
		if ( qnt <= 0 )
			qnt = 1;
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(ph);
		MidiPhrase r = new MidiPhrase();
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			int rem = (int)((m.getTime()) % qnt);
			long delta;
			if ( (rem*2) <= qnt )
				delta = -rem;
			else
				delta = (qnt-rem);
			m.setTime(m.getTime()+delta);
			r.merge(m);
		}
		return(r);
	}

	/** Try to put all notes within the limits of pitches p1,p2
	 * This is accomplished by shifting notes in octave increments.
	 */
	public static MidiPhrase octaveFence(MidiPhrase ph,int p1,int p2) {
		ph = Alg.onlyNotes(ph);
		int pmid = (p1+p2)/2;
	
		int php1 = lowest(ph);
		int php2 = highest(ph);
		int phmid = (php1+php2)/2;
	
		// First shift the whole phrase up or down to get everything
		// as close as possible while retaining relative melodies.
		ph = Alg.transpose(ph, (int)numquant(pmid-phmid,12) );
	
		// Then move individual notes 
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(ph);
		MidiPhrase r = new MidiPhrase();
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			m = (MidiMsg)m.clone();		// since we're going to modify it
			while ( m.getPitch() < p1 )
				m.setPitch(m.getPitch()+12);
			while ( m.getPitch() > p2 )
				m.setPitch(m.getPitch()-12);
			r.merge(m);
		}
		return(r);
	}

	/**	Scale a phrase by multiplying the time and duration
	 * each note by a specified factor.
	 */
	public static MidiPhrase scaTimes(MidiPhrase ph, double f) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(ph);
		MidiPhrase r = new MidiPhrase();
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			m = (MidiMsg)m.clone();
			m.setTime((long)(m.getTime() * f));
			m.setDuration((long)(m.getDuration() * f));
			r.merge(m);
		}
		r.setLength((long)(ph.getLength() * f));
		return(ph);
	}

	public static MidiPhrase[] split(MidiPhrase p) {
		MidiMsg n;
		MidiMsg newn;
		MidiPhrase p2;
		long tm2;
		boolean samesection;
		long t, elapse, closest;
		int nactive;
		long now;
		MidiMsg[] activent;
		long[] activetime;
		MidiPhrase[] arr;
		int MAXSIMUL = 64;

		nactive = 0;
		now = 0L;
		activent = new MidiMsg[MAXSIMUL];
		activetime = new long[MAXSIMUL];
		arr = new MidiPhrase[0];

		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);

		n = (MidiMsg)e.nextElement();

		while ( n!=null || nactive>0 ) {
	
			// find out which event is closer: the end of a pending
			// note, or the start of the next one (if there is one).
	
			if ( n != null ) {
				closest = n.getTime() - now;
				samesection = true;
			}
			else {
				closest = 30000;
				samesection = false;
			}
	
			// Check the ending times of the pending notes, to see
			// if any of them end before the next note starts.  If so,
			// we want to create a new section.
			for ( int k=0; k<nactive; k++ ) {
				if ( (t=activetime[k]) <= closest ) {
					closest = t;
					samesection = false;
				}
			}
	
			// We want to let that amount of time elapse.
			elapse = closest;
			if ( ! (samesection && (closest == 0 || nactive == 0)) ) {
	
				// We're going to create a new element in the split array
		
				MidiPhrase[] newarr = new MidiPhrase[arr.length+1];
				System.arraycopy(arr,0,newarr,0,arr.length);
				p2 = new MidiPhrase();
				newarr[arr.length] = p2;
				arr = newarr;

				// add all active notes to the phrase
				tm2 = now+elapse;
				for ( int k=0; k<nactive; k++ ) {
					newn = (MidiMsg)activent[k].clone();
					if ( newn.getTime() < now ) {
						long overhang = now - newn.getTime();
						newn.setTime(newn.getTime() + overhang);
						newn.setDuration(newn.getDuration() - overhang);
					}
					long endof = newn.getTime() + newn.getDuration();
					if ( endof > tm2 ) {
						newn.setDuration(tm2-newn.getTime());
					}
					p2.merge(newn);
				}
				p2.setLength(tm2);
		
				// If any notes are pending, take into account elapsed
				// time, and if they expire, get rid of them.
				for ( int i=0; i<nactive; i++ ) {
					if ( (activetime[i] -= elapse) <= 0L ) {
						// Remove this note from the list by
						// shifting everything down.
						for ( int j=i+1; j<nactive; j++ ) {
							activetime[j-1] = activetime[j];
							activent[j-1] = activent[j];
						}
						nactive--;
						i--;	// don't advance loop index
					}
				}
			}

			if ( samesection ) {
				// add this new note (and all others that start at the
				// same time) to the list of active ones
				long thistime = n.getTime();
				while ( n!=null && n.getTime() == thistime ) {
					activent[nactive] = n;
					activetime[nactive++] = n.getDuration();
					// advance to next note
					n = (MidiMsg)e.nextElement();
				}
			}
			now += elapse;
		}
		return arr;
	}

	/** Returns a monophonic version of the specified phrase.
	 * If type==0, high notes are given priority (e.g. if two notes are
	 * playing simultaneously, the higher note will be chosen).
	 * If type==1, low notes are given priority.  If type==2,
	 * the priority is randomized.
	 */

	public static MidiPhrase mono(MidiPhrase ph,int typ) {
		MidiPhrase non = nonNotes(ph);
		ph = onlyNotes(ph);
		MidiPhrase[] arr = split(ph);
		MidiPhrase r = new MidiPhrase();
		int n = arr.length;
		if ( typ == 1 ) {
			while ( --n >= 0 ) {
				r.merge(lowestnt(arr[n]));
			}
		}
		else if ( typ == 2 ) {
			Random rand = new Random();
			while ( --n >= 0 ) {
				// pick a random one from each section
				int asize = arr[n].size();
				int rnum = rand.nextInt();
				if ( rnum < 0 )
					rnum = -rnum;
				int rn = rnum%asize;

// System.out.println("asize="+Integer.toString(asize)
// 			+" rnum="+Integer.toString(rnum)
// 			+" run="+Integer.toString(rn));
// System.out.println("ph="+(arr[n].getNth(1+rn)).toString());

				r.merge(arr[n].getNth(1+rn));
			}
		}
		else {
			while ( --n >= 0 ) {
				r.merge(highestnt(arr[n]));
			}
		}
		r.merge(non);
		return(r);
	}

	/** Return a phrase containing every Nth note of a phrase */
	public static MidiPhrase everyNth(MidiPhrase p, int n) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		MidiPhrase r = new MidiPhrase();
		for ( int num=1; e.hasMoreElements(); num++ ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			if ( (num%n) == 0 ) {
				r.merge(m);
			}
		}
		return r;
	}
	
	/** Return a phrase containing a random selection of notes,
	 * choosing approximately 1 of every n notes.
	 */
	public static MidiPhrase everyRandomNth(MidiPhrase p, int n) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		MidiPhrase r = new MidiPhrase();
		Random rnd = new Random();
		for ( int num=1; e.hasMoreElements(); num++ ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			int i = rnd.nextInt();
			if ( i < 0 )
				i = -i;
			if ( (i%n) == 0 )
				r.merge(m);
		}
		return r;
	}
	
	/** Return a phrase containing notes that are on a particular beat. */
	public static MidiPhrase onBeat(MidiPhrase p, long d) {
		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		MidiPhrase r = new MidiPhrase();
System.out.println("onBeat needs an implementation!");
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			r.merge(m);
		}
		return r;
	}

	/** Return a program-change message.
	 * patchnum is 1-based, ch (channel) is 1-based
	 */
	public static MidiPhrase progchange(int patchnum,int ch) {
		return(new MidiPhrase(new MidiBytes((byte)(0xc0|(ch-1)), (byte)(patchnum-1))));
	}

	public static MidiPhrase gvolchange(int v) {
		MidiPhrase p = new MidiPhrase();
		for ( int c=0; c<16; c++ ) {
			p.append(new MidiBytes((byte)(0xb0|c),(byte)0x07,(byte)v));
		}
		return p;
	}

	/** Generate a phrase with controller messages that press the sustain
	 * pedal on the specified channel (c) and let it up after the specified
	 * amount of time (tm).
	 */
	public static MidiPhrase sussect(long tm, int c) {
		int b = (int)MidiSystem.clicksPerBeat();
		MidiBytes s1 = new MidiBytes((byte)(0xb0|(c-1)),
			(byte)0x40,(byte)0x7f); 	// sustain on
		MidiBytes s2 = new MidiBytes((byte)(0xb0|(c-1)),
			(byte)0x40,(byte)0x00); 	// sustain off
		s2.setTime(tm-1*b/8);		// -1b/8 so it won't overlap next section
		MidiPhrase s = new MidiPhrase(s1);
		s.merge(new MidiPhrase(s2));
		s.setLength(tm);
		return(s);
	}
	
	/** Return a phrase of notes based on the letters in a String */
	public static MidiPhrase strnotes(String s) {
		int lng = s.length();
		MidiPhrase r = new MidiPhrase();
		for ( int n=0; n<lng; n++ ) {
			int a = s.charAt(n) - 'a';
			if ( a < 0 )
				a = -a;
			MidiMsg nt = new MidiNote(48 + a);
			nt.setDuration((1+a/12)*MidiSystem.clicksPerBeat()/2);
			r.append(new MidiPhrase(nt));
		}
		return(r);
	}

	public static void main(String args[]) {

		long t1, t2;
		t1 = System.currentTimeMillis();

		MidiPhrase p = Alg.strnotes("abc");
System.out.println("p size = ="+p.size());
		for ( int k = 500; k<10000; k+=500 ) {
			t1 = System.currentTimeMillis();
			p = Alg.strnotes("abc");
			p = Alg.repeat(p,k);
			t2 = System.currentTimeMillis();
System.out.println("p size = ="+p.size());
System.out.println("time="+t2+"  diff="+(t2-t1));
		}

		t2 = System.currentTimeMillis();
System.out.println("Matic end time="+t2+"  diff="+(t2-t1));

/*
		System.out.println("main start");
		MidiPhrase p = new MidiPhrase();
		p.append(new MidiNote(100));
		p.append(new MidiNote(101));
		p.append(new MidiBytes((byte)1,(byte)2,(byte)3));
		System.out.println("p="+p.toString());
		System.out.println("nonNotes="+nonNotes(p).toString());
		System.out.println("onlyNotes="+onlyNotes(p).toString());

		MidiPhrase p2 = new MidiPhrase();
		p2.append(new MidiNote(101));
		p2.append(new MidiNote(102));
		p2.append(new MidiNote(100));
		p2.setTime(0);
		System.out.println("p2="+p2.toString());
		p2.reOrder();
		System.out.println("p2 reOrdered="+p2.toString());

		System.out.println("findRoot="+findRoot(p).toString());
		System.out.println("main end");
*/
	}
}

