/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * Manage MIDI messages that consist of just bytes (e.g. sysex).
 *
 * @author Tim Thompson
 */

public class MidiBytes extends MidiMsg implements Cloneable {

	// Should eventually change this code to allocate new space
	// in chunks, rather than allocating a new array for every new byte.

	private byte[] bytes;

	/**
	 * Create a MidiBytes message.
	 * @param t time (in clicks) of the message.
	 * @param b array of bytes to be stored in the message.
	 */
	public MidiBytes(long t, byte[] b) {
		super(t);
		bytes = b;
	}
	/**
	 * Create a MidiBytes message.
	 * The default time of the message is is 0.
	 */
	public MidiBytes() {
		this(0,new byte[0]);
	}

	/**
	 * Create a MidiBytes message.
	 * The message is initialized with a single byte.
	 */
	public MidiBytes(byte c1) {
		super(0);
		bytes = new byte[1];
		bytes[0] = c1;
	}
	/**
	 * Create a MidiBytes message.
	 * The message is initialized with two bytes.
	 */
	public MidiBytes(byte c1, byte c2) {
		super(0);
		bytes = new byte[2];
		bytes[0] = c1;
		bytes[1] = c2;
	}
	/**
	 * Create a MidiBytes message.
	 * The message is initialized with three bytes.
	 */
	public MidiBytes(byte c1, byte c2, byte c3) {
		super(0);
		bytes = new byte[3];
		bytes[0] = c1;
		bytes[1] = c2;
		bytes[2] = c3;
	}

	public Object clone() {
		MidiBytes mb = new MidiBytes();
		mb.time = this.time;
		mb.next = null;		// needed?
		mb.bytes = new byte[bytes.length];
		for ( int n=0; n<bytes.length; n++ )
			mb.bytes[n] = bytes[n];
		return mb;
	}

	public boolean equals(MidiMsg m) {
		if ( ! super.equals(m) )
			return false;
		if ( ! (m instanceof MidiBytes) )
			return false;
		MidiBytes mb = (MidiBytes)m;
		if ( ! mb.bytes.equals(bytes) )
			return false;
		return true;
	}

	/**
	 * Return the pitch of this message.
	 * For MidiByte messages, this is meaningless, so -1 is returned.
	 * This method (and the next few methods) are included so that
	 * code can be written conveniently using just the MidiMsg methods.
	 */
	public int getPitch() { return -1; }

	/**
	 * Set the pitch of this message.
	 * For MidiByte messages, this is meaningless, so it is ignored.
	 */
	public void setPitch(int p) { }

	/**
	 * Return the duration of this message.
	 * For MidiByte messages, this is meaningless, so -1 is returned.
	 */
	public long getDuration() { return 0; }	// keep this 0, for append()

	/**
	 * Set the duration of this message.
	 * For MidiByte messages, this is meaningless, so it is ignored.
	 */
	public void setDuration(long d) { }

	/**
	 * Return the channel of this message.
	 * For MidiBytes messages, this is is potentially useful (e.g. for
	 * controller messages and things), but for the moment we return -1.
	 */
	public int getChannel() { return -1; }

	/**
	 * Set the channel of this message.
	 * For MidiBytes messages, this is is potentially useful (e.g. for
	 * controller messages and things), but for the moment we ignore it.
	 */
	public void setChannel(int c) { }

	/**
	 * Return the velocity of this message.
	 * For MidiBytes messages, this is meaningless, so we return -1.
	 */
	public int getVelocity() { return -1; }

	/**
	 * Set the velocity of this message.
	 * For MidiBytes messages, this is meaningless, so we ignore it.
	 */
	public void setVelocity(int v) { }

	/**
	 * Return an array of bytes containing the contents of the message.
	 */
	public byte[] getBytesOn() { return bytes; }

	/**
	 * Return an array of bytes containing the ending half of the message.
	 * For MidiBytes messages, this is always null.   In the future, though,
	 * there might be things other than MidiNote messages (for example,
	 * like MidiSustain messages) that have duration, and which would 
	 * make use of getBytesOff().
	 */
	public byte[] getBytesOff() { return null; }

	/**
	 * Add a byte to the end of this message. 
	 */
	public void addByte(byte b) {
		int oldlength = bytes.length;
		byte[] newbytes = new byte[oldlength+1];
		System.arraycopy(bytes,0,newbytes,0,oldlength);
		newbytes[oldlength] = b;
		bytes = newbytes;
	}

	/**
	 * Return a human-readable representation of the message.
	 */
	public String toString() {
		String s = "MidiBytes(t="+Long.toString(time)+",";
		for ( int i=0; i<bytes.length; i++ ) {
			if ( i > 0 )
				s += ",";
			s += Integer.toString(bytes[i]);
		}
		s += ")";
		return s;
	}
}
