/* NoSuch MIDI - Copyright 1997 by Tim Thompson and NoSuch Media */

package com.nosuch.midi;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Write a Standard MIDI File.
 *
 * <pre>
 * Typical usage:
 *   FileOutputStream fos = new FileOutputStream("test.mid");
 *   MidiFileWriter mf = new MidiFileWriter(outputstream);
 *   mf.addTrack(new MidiPhrase('a,b,c,d,e'));
 *   mf.write();
 *   fos.close();
 * </pre>
 */

public class MidiFileWriter {

	OutputStream _out;
	int _fmt;
	int _div;
	Vector _tracks;
	int _tempo;
	ByteArrayOutputStream _track;		// while being constructed
	double _clickfactor = 1.0;
	int _status;
	static final int DEFCLICKS = 96;

	static final int NOTEON = 0x90;
	static final int NOTEOFF = 0x80;
	static final int AFTERTOUCH = 0xA0;
	static final int CONTROLLER = 0xB0;
	static final int PROGRAMCHANGE = 0xC0;
	static final int CHANAFTERTOUCH = 0xD0;
	static final int PITCHBEND = 0xE0;
	static final int SYSEX = 0xF0;
	static final int SYSEXEND = 0xF7;

	/**
	 * Create a MidiFileWriter for sending a Standard MIDI File
	 * to an OutputStream.  Defaults to a type-0 file with no tempo track.
	 * @param	out	The OutputStream to send data to.
	 */
	public MidiFileWriter(OutputStream out) {
		this(out,0,-1,DEFCLICKS);
	}

	/**
	 * Create a MidiFileWriter for sending a Standard MIDI File
	 * to an OutputStream.
	 * @param	out	The OutputStream to send data to.
	 * @param	f	The Standard MIDI File type (0, 1, or 2).
	 * @param	t	The tempo, in microseconds.  This is only used
	 *                      if the file type is 1, in which case a tempo
	 *                      track is added (assuming this value is > 0)
	 */
	public MidiFileWriter(OutputStream out, int f, int t) {
		this(out,f,t,DEFCLICKS);
	}

	/**
	 * Create a MidiFileWriter for sending a Standard MIDI File
	 * to an OutputStream.
	 * @param	out	The OutputStream to send data to.
	 * @param	f	The Standard MIDI File type (0, 1, or 2).
	 * @param	t	The tempo, in microseconds.  This is only used
	 *                      if the file type is 1, in which case a tempo
	 *                      track is added (assuming this value is > 0)
	 * @param	d	Number of clicks per quarter note?
	 */
	public MidiFileWriter(OutputStream out, int f, int t, int d) {
		_out = out;
		_tracks = new Vector();
		_fmt = f;
		_div = d;
		_clickfactor = _div / DEFCLICKS;
		_tempo = t;
		_track = new ByteArrayOutputStream();
	}

	/**
	 * Add a track to the list of tracks that will be written.
	 * (Nothing is actually written until the write() method is called.
	 * @param	p	The MidiPhrase to be added.
	 */
	public void addTrack(MidiPhrase p) {
		_tracks.addElement(p);
	}

	/**
	 * Write the Midi File.
	 */
	public void write() throws IOException {

		output_MThd();

		int ntracks = _tracks.size();
		if ( _fmt == 1 && _tempo > 0 )
			ntracks++;

		output_32bit(6);
		output_16bit(_fmt);
		output_16bit(ntracks);
		output_16bit(_div);
		if ( _fmt == 1 && _tempo > 0 ) {
			tempo_track(_tempo);
			output_track();
		}
		for ( Enumeration e=_tracks.elements(); e.hasMoreElements(); ){
			MidiPhrase p = (MidiPhrase) e.nextElement();
			phrase_track(p);
			output_track();
		}
		_out.flush();
	}

	private void output_MThd() throws IOException {
		byte[] hdr = { 'M', 'T', 'h', 'd' };
		_out.write(hdr);
	}

	private void output_MTrk() throws IOException {
		byte[] hdr = { 'M', 'T', 'r', 'k' };
		_out.write(hdr);
	}

	private void output_track() throws IOException {
		// send _track that has been collected, to output stream _out
		_track.flush();
		output_MTrk();
		output_32bit(_track.size());
		_track.writeTo(_out);
		_track.reset();
	}

	private void output_32bit(int n) throws IOException {
		byte[] bt = new byte[4];
		bt[0] = (byte) (n>>24 & 0xff);
		bt[1] = (byte) (n>>16 & 0xff);
		bt[2] = (byte) (n>>8 & 0xff);
		bt[3] = (byte) (n & 0xff);
		_out.write(bt);
	}

	private void output_16bit(int n) throws IOException {
		byte[] bt = new byte[2];
		bt[0] = (byte) (n>>8 & 0xff);
		bt[1] = (byte) (n & 0xff);
		_out.write(bt);
	}

	// These methods deal with the real output steam.

	private void tempo_track(int tempo) throws IOException {
		inittrack();
		putdelta(0);
		timesig(4,2,24,8);
		putdelta(0);
		mftempo(tempo);
		putdelta(0);
		endOfTrack();
	}

	private void phrase_track(MidiPhrase p) throws IOException {
		inittrack();
		int tm = 0;
		MidiPhrase pendingOffs = new MidiPhrase();
		// We want a copy of p, so we can mess with its MidiMsgs
		p = (MidiPhrase) p.clone();
		MidiMsg topMsg = p.removeFirstMsg();
		while ( topMsg!=null || pendingOffs.peekFirstMsg()!=null  ) {

			// This must be done each time, because pendingOffs
			// is changing.
			MidiMsg topPendingMsg = pendingOffs.peekFirstMsg();

			if ( topMsg==null || (topPendingMsg!=null && topPendingMsg.getTime() <= topMsg.getTime() ) ) {

				// output pendingOffs
				int t = (int) topPendingMsg.getTime();
				putdelta(t-tm);
				tm = t;
				int ch = topPendingMsg.getChannel()-1;
				putnote(NOTEON | ch,topPendingMsg.getPitch(),0);
				// advance to next note in pending list
				pendingOffs.removeFirstMsg();
			}
			else {
				int t = (int) topMsg.getTime();

				// MidiMsg channels are 1-based
				int ch = topMsg.getChannel()-1;
				int pit = topMsg.getPitch();
				int vel = topMsg.getVelocity();

				putdelta(t-tm);
				tm = t;
				if ( topMsg instanceof MidiNoteOn ) {
					putnote(NOTEON | ch,pit,vel);
				}
				else if ( topMsg instanceof MidiNoteOff ) {
					putnote(NOTEOFF | ch,pit,vel);
				}
				else if ( topMsg instanceof MidiBytes ) {
					putbytes(topMsg);
				}
				else if ( topMsg instanceof MidiNote ) {
					putnote(NOTEON | ch,pit,vel);
					// Move it to list of notes for which
					// pending note-offs are needed
					topMsg.setTime(t+topMsg.getDuration());
					pendingOffs.merge(topMsg);
				}
				else {
					throw new IOException("Unrecognized MidiMsg type!?");
				}
				// advance to next note in main list
				topMsg = p.removeFirstMsg();
			}
		}
		putdelta(0);
		endOfTrack();
	}

	private void putnote(int c, int p, int v) throws IOException {
		if ( c != _status ) {
			trackbyte(c);
			_status = c;
		}
		trackbyte(p);
		trackbyte(v);
	}

	private void putbytes(MidiMsg n) throws IOException {
		byte[] b = n.getBytesOn();
		int byte1 = b[0];
		int togo = -1;
		int bt = byte1 & 0xf0;
		switch ( bt ) {
		case NOTEON:
		case NOTEOFF:
		case AFTERTOUCH:
		case CONTROLLER:
		case PITCHBEND:
			togo = 3;
			break;
		case PROGRAMCHANGE:
		case CHANAFTERTOUCH:
			togo = 2;
			break;
		}
		int i;
		if ( b.length == togo ) {
			if ( byte1 != _status ) {
				trackbyte(byte1);
				_status = byte1;
			}
			for ( i=1; i<togo; i++ )
				trackbyte(b[i]);
		}
		else {
			trackbyte(0xf7);
			togo = b.length;
			writevarinum(togo);
			for ( i=0; i<togo; i++ )
				trackbyte(b[i]);
			_status = b[0];
		}
	}

	private void trackbyte(int c) throws IOException {
		_track.write(c);
	}

	private void putdelta(int click) throws IOException {
		if ( click < 0 )
			throw new IOException("Negative click value");
		int t = (int)(click*_clickfactor + 0.5);
		writevarinum(t);
	}

	private void timesig(int n, int denom, int c, int b) throws IOException {
		trackbyte(0xff);
		trackbyte(0x58);
		trackbyte(0x04);
		trackbyte(n);
		trackbyte(denom);
		trackbyte(c);
		trackbyte(b);
	}

	private void mftempo(int microsecs) throws IOException {
		trackbyte(0xff);
		trackbyte(0x51);
		trackbyte(0x03);
		trackbyte( (microsecs>>16) & 0xff );
		trackbyte( (microsecs>>8) & 0xff );
		trackbyte( microsecs & 0xff );
	}

	private void writevarinum(int value) throws IOException {
		int buffer = value & 0x7f;
		
		while ( (value >>= 7) > 0 ) {
			buffer <<= 8;
			buffer |= 0x80;
			buffer += (value & 0x7f);
		}
		for ( ;; ) {
			trackbyte((int)(buffer&0xff));
			if ( (buffer & 0x80) == 0 )
				break;
			buffer >>= 8;
		}
	}

	private void inittrack() {
		_status = -1;
		_track.reset();
	}

	private void endOfTrack() throws IOException {
		trackbyte(0xff);
		trackbyte(0x2f);
		trackbyte(0x00);
	}

	public static void main(String[] args) {

		if ( args.length < 2 ) {
			System.out.println("usage: midifilewriter {phrase} {file}");
			System.exit(1);
		}
		String ps = args[0];
		String fname = args[1];

		FileOutputStream o = null;
		try {
			o = new FileOutputStream(fname);
		} catch (IOException e) {
			System.out.println("Exception opening "+fname+", msg="+e.getMessage());
			System.exit(1);
		}
		// MidiFileWriter mf = new MidiFileWriter(o,1,500000);
		MidiFileWriter mf = new MidiFileWriter(o);
		MidiPhrase p = new MidiPhrase(ps);
		System.out.println("Phrase is "+p);
		
		mf.addTrack(p);
		try {
			mf.write();
		} catch (IOException e) {
			System.out.println("Exception writing "+fname+", msg="+e.getMessage());
			System.exit(1);
		}
	}

}
