/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * Finite State Machine for parsing incoming MIDI bytes.
 *
 * @author Tim Thompson
 */

public class MidiFsm {

	private final int OFF = 1;
	private final int ON = 2;
	private final int PRESSURE = 3;
	private final int CONTROLLER = 4;
	private final int PROGRAM = 5;
	private final int CHANPRESS = 6;
	private final int PITCHBEND = 7;
	private final int SYSEX = 8;
	private final int POSITION = 9;
	private final int SONG = 10;
	private final int TUNE = 11;
	private final int EOX = 12;

	private final int NEEDSTATUS = 0;
	private final int NEEDONEBYTE = 1;
	private final int NEEDTWOBYTES = 2;
	private final int VARIABLELENGTH = 3;

	private int state = NEEDSTATUS;
	private int chan = 0;
	private byte mess[];
	private int messindex = 0;
	private int func = 0;

	private boolean runnable = false;	// if true, the current message
										// can use "running status byte".
	private MidiAction act;

	/**
	 * Construct a new Fsm.  The MidiAction object receives
	 * method invocations as the MIDI messages are recognized
	 * @param a MidiAction object whose methods are invoked as callbacks.
	 */
	public MidiFsm(MidiAction a) {
		act = a;
		mess = new byte[0];
		parse(-1);		// to reset
	}

	/**
	 * Parse a single MIDI byte.
	 * If it completes a MIDI message, the callback methods of the
	 * MidiAction object are invoked.  Calling parse(-1) will reset
	 * the state machine.
	 * @param inbyte Input MIDI byte (or -1 to reset state machine).
	 */
	public void parse(int inbyte) {

		boolean highbit;

		if ( messindex >= mess.length ) {
			byte newmess[] = new byte[messindex + 64];
			System.arraycopy(mess,0,newmess,0,mess.length);
			mess = newmess;
		}
		if ( inbyte == -1 ) {
			messindex = 0;
			mess[messindex++] = (byte)0x90;
			chanmsg(0x90);
			return;
		}
		if ( (inbyte & 0x80) != 0 )
			highbit = true;
		else
			highbit = false;

		// Check active sensing early, so it doesn't interrupt sysex.
		if ( inbyte == 0xfe ) {
			// do nothing with active sensing
			return;
		}
	
		// Variable length messages (ie. the 'system exclusive')
		// can be terminated by the next status byte, not necessarily
		// an EOX.  Actually, since EOX is a status byte, this
		// code ALWAYS handles the end of a VARIABLELENGTH message.

		if ( state==VARIABLELENGTH && highbit==true )  {
			// The message has ended; call func and reset.
			if ( func>0 && messindex>0 )
				dofunc(func);
			func = 0;
			state = NEEDSTATUS;
		}
	
		// Real time messages can occur ANYPLACE,
		// but do not interrupt running status.

		if ( (inbyte & 0xf8) == 0xf8 ) {
			realmess(inbyte);
			return;
		}
	
		// Status bytes always start a new message.
	
		if ( highbit ) {
			messindex = 0;
			mess[messindex++] = (byte)inbyte;
			if ( (inbyte & 0xf0) == 0xf0 ) {
				sysmess(inbyte);
				runnable = false;
			}
			else
				chanmsg(inbyte);
			return;
		}
	
		// We've got a Data byte.
	
		mess[messindex++] = (byte)inbyte;
	
		switch ( state ) {
		case NEEDSTATUS:
			// We shouldn't get here, since in NEEDSTATUS mode
			// we're expecting a new status byte, NOT any
			// data bytes.
			System.err.println("Got a data byte ("+
				Integer.toString(inbyte)+") while expecting a status byte!");
			break;
		case NEEDTWOBYTES:
			// wait for the second byte
			if ( messindex < 3 )
				return;
			//FALLTHRU!!
		case NEEDONEBYTE:
			// We've completed a 1 or 2 byte message.
			if ( func > 0 )
				dofunc(func);
			if ( runnable ) {
				// In runnable mode, we reset the message
				// index, but keep the func and state the
				// same.  This provides the "running status
				// byte" feature.
				messindex = 1;
			}
			else {
				// If not runnable, reset to NEEDSTATUS mode
				state = NEEDSTATUS;
				func = 0;
			}
			break;
		case VARIABLELENGTH:
			// nothing to do
			break;
		}
		return;
	}
	
	// realmess(inbyte)
	//
	// Call the real-time function for the specified byte, immediately.
	// These can occur anywhere, so they don't change the state.
	
	private void realmess(int inbyte) {
	
		if ( act == null )
			return;

		switch ( inbyte ) {
		case 0xf8:
			act.timing();
			break;
		case 0xfa:
			act.start();
			break;
		case 0xfb:
			act.midiContinue();
			break;
		case 0xfc:
			act.stop();
			break;
		case 0xfe:
			// do nothing, active sensing gets handled earlier
			break;
		case 0xff:
			act.reset();
			break;
		}
		return;
	}
	
	// chanmsg(inbyte)
	//
	// Interpret a Channel (voice or mode) Message status byte.
	
	private void chanmsg(int inbyte) {

		chan = inbyte & 0xf;// channel # is lower 4 bits
		chan++;				// Channel we give to action routines is 1-based
		runnable = true;		// Channel messages can use running status
	
		// The high 4 bits, which determine the type of channel message.
	
		switch ( inbyte & 0xf0 ) {
		case 0x80:
			func = OFF;
			state = NEEDTWOBYTES;
			break;
		case 0x90:
			func = ON;
			state = NEEDTWOBYTES;
			break;
		case 0xa0:
			func = PRESSURE;
			state = NEEDTWOBYTES;
			break;
		case 0xb0:
			func = CONTROLLER;
			state = NEEDTWOBYTES;
			break;
		case 0xc0:
			func = PROGRAM;
			state = NEEDONEBYTE;
			break;
		case 0xd0:
			func = CHANPRESS;
			state = NEEDONEBYTE;
			break;
		case 0xe0:
			func = PITCHBEND;
			state = NEEDTWOBYTES;
			break;
		}
		return;
	}
	
	// sysmess(inbyte)
	//
	// Initialize (and possibly call) the MIDI function for the specified byte.
	// Set the state that the state-machine should go into.  If the function
	// is not called immediately, func is set, so that when the state machine
	// gets to the end of the MIDI message, it can then call the function.
	// If the function IS called immediately, the finite state machine is reset
	// (func=NULL,state=NEEDSTATUS).
	
	private void sysmess(int inbyte) {

		int f = 0;
	
		switch ( inbyte ) {
		case 0xf0:
			func = SYSEX;
			state = VARIABLELENGTH;
			break;
		case 0xf2:
			func = POSITION;
			state = NEEDTWOBYTES;
			break;
		case 0xf3:
			func = SONG;
			state = NEEDONEBYTE;
			break;
		case 0xf6:
			f = TUNE;
			func = 0;
			state = NEEDSTATUS;
			break;
		case 0xf7:
			f = EOX;
			func = 0;
			state = NEEDSTATUS;
			break;
		}
		if ( f > 0 )
			dofunc(f);
		return;
	}
	private void dofunc(int f) {

		if ( act == null )
			return;

		// The channel numbers we send are 1-based.

		switch (f) {
		case OFF:
			act.off(chan,mess[1],mess[2]);
			break;
		case ON:
			act.on(chan,mess[1],mess[2]);
			break;
		case PRESSURE:
			act.pressure(chan,mess[1],mess[2]);
			break;
		case CONTROLLER:
			act.controller(chan,mess[1],mess[2]);
			break;
		case PROGRAM:
			act.program(chan,mess[1]);
			break;
		case CHANPRESS:
			act.channelPressure(chan,mess[1]);
			break;
		case PITCHBEND:	
			act.pitchBend(chan,mess[1],mess[2]);
			break;
		case SYSEX:
			byte[] sx;
			sx = new byte[messindex+1];
			System.arraycopy(mess,0,sx,0,messindex);
			act.sysex(sx);
			break;
		case POSITION:
			act.position(mess[1],mess[2]);
			break;
		case SONG:
			act.song(mess[1]);
			break;
		case TUNE:
			act.tune();
			break;
		case EOX:
			act.eox();
			break;
		default:
			System.err.println("Unrecognized value in MidiFsm.dofunc! f="+Integer.toString(f));
			break;
		}
	}
}

