/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * MidiMsg abstract class.  Used for both "notes" (whether
 * they are complete notes or just isolated NoteOn's or NotOff's)
 * as well as MIDI system exclusives and controller messages.
 *
 * The fundamental attribute of a MidiMsg is its time.
 * MidiPhrase objects are lists of MidiMsg objects, ordered by time.
 *
 * Some of the methods (like getPitch/setPitch) are only applicable
 * to "note-like" objects.  For non-note-like objects, these are no-ops.
 *
 * @author	Tim Thompson
 */

public abstract class MidiMsg implements Cloneable {

	protected long time;
	protected MidiMsg next;

	public MidiMsg() {
		time = 0;
		next = null;
	}
	public MidiMsg(long tm) {
		time = tm;
		next = null;
	}

	public void setTime(long tm) {
		time = tm;
	}
	public long getTime() {
		return time;
	}

	/**
	 * Get the pitch (0 to 127).
	 */
	abstract public int getPitch();
	/**
	 * Set the pitch (0 to 127).
	 */
	abstract public void setPitch(int p);

	/**
	 * Get the duration (in clicks).
	 */
	abstract public long getDuration();
	/**
	 * Set the duration (in clicks).
	 */
	abstract public void setDuration(long d);

	/**
	 * Get the channel (1 to 16).
	 */
	abstract public int getChannel();
	/**
	 * Set the channel (1 to 16).
	 */
	abstract public void setChannel(int c);

	/**
	 * Get the velocity (0 to 127).
	 */
	abstract public int getVelocity();
	/**
	 * Set the velocity (0 to 127).
	 */
	abstract public void setVelocity(int v);

	/**
	 * Get the array of bytes representing the start of the message.
	 */
	abstract public byte[] getBytesOn();

	/**
	 * Get the array of bytes representing the end of the message.
	 * This method will return non-null for things that have duration,
	 * such as a MidiNote.
	 */
	abstract public byte[] getBytesOff();

	/**
	 * Return a human-readable version.
	 */
	abstract public String toString();

	/**
	 * Return a new copy.
	 */
	abstract public Object clone();

	/**
	 * Return true if this object equals another.
	 */
	public boolean equals(MidiMsg m) {
		boolean b = (m.getTime() == time);
		return b;
	}

	/**
	 * Return true if this MidiMsg is less than the MidiMsg parameter,
	 * in terms of ordering criteria within a MidiPhrase.  This method
	 * should NOT be used for determining whether two MidiMsgs are "equal".
	 * @param m MidiMsg to compare
	 */
	public int compareOrder(MidiMsg m2) {
		long ld = time - m2.time;
		if ( ld < 0 )
			return -1;
		if ( ld > 0 )
			return 1;
		// Time is identical.  If they're Noteable notes, compare pitch
		if ( this instanceof MidiNoteable && m2 instanceof MidiNoteable ) {
			int d = this.getPitch() - m2.getPitch();
			if ( d < 0 )
				return -1;
			if ( d > 0 )
				return 1;
			int c = this.getChannel() - m2.getChannel();
			if ( c < 0 )
				return -1;
			if ( c > 0 )
				return 1;
			int v = this.getVelocity() - m2.getVelocity();
			if ( v < 0 )
				return -1;
			if ( v > 0 )
				return 1;
			return 0;
		}
		return 0;
	}
}
