/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

import java.util.*;
import java.io.*;

/**
 * A MidiPhrase is an ordered (by time) list of MidiMsg objects.
 * A MidiPhrase also has a "length" attribute that is the number
 * of clicks from its beginning to its end.  This length is used
 * when concatenating phrases, among other things.  The length is
 * completely independent of the MidiMsg objects within the phrase,
 * allowing for phrases with leading space, or phrases that are
 * entirely space.
 *
 * @version 0.0 1/1/96
 * @author	Tim Thompson
*/

public class MidiPhrase implements Cloneable, Serializable {

	protected MidiMsg _msg;
	private long _phraseLength;
	protected MidiMsg _lastmsg;	// last note in phrase

	// Default values saved while parsing KeyKit-style constructors
	private int Defvol;
	private int Defoct;
	private long Defdur;
	private int Defchan;
	private long Deftime;
	private long Def2time;

	/** Construct a new phrase with a specified MidiMsg and length
	 * @param n MidiMsg
	 * @param l length of phrase (in clicks)
	 */
	public MidiPhrase(MidiMsg n,long l) {
		_msg = n;
		_lastmsg = n;
		_phraseLength = l;
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.writeUTF(this.toString());
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		String s = in.readUTF();
		mergeKeyKitValue(s);
	}

	/** Construct a new phrase with a specified MidiMsg
	 * @param n MidiMsg
	 */
	public MidiPhrase(MidiMsg n) {
		this(n,n.getDuration());
	}

	/** Construct a new phrase with no MidiMsgs and 0 length. */
	public MidiPhrase() {
		this(null,0);
	}

	/** Construct a new phrase from a KeyKit-style constant string. */
	public MidiPhrase(String s) {
		this(null,0);
		mergeKeyKitValue(s);
	}

	/** Return the current length of the phrase (in clicks). */
	public long getLength() {
		return _phraseLength;
	}

	/** Return an enumeration of the MidiMsgs in the phrase. */
	public Enumeration getEnumeration() {
		return new MidiPhraseEnumeration(this);
	}

	/** Return the number of MidiMsgs in the phrase. */
	public int size() {
		int n = 0;
		for ( MidiMsg m=_msg; m!=null; m=m.next )
			n++;
		return n;
	}

	/** Set the length of the phrase (in clicks).
	 * @param l length (in clicks)
	 */
	public MidiPhrase setLength(long l) {
		_phraseLength = l;
		return this;
	}

	/** Return the first message in the phrase */
	public boolean isEmpty() {
		if ( _msg == null )
			return true;
		else
			return false;
	}

	/** Return the first message in the phrase.  The name is peek* so that
	 * people won't be tempted to actually modify it. */
	public MidiMsg peekFirstMsg() {
		return _msg;
	}

	/** Return and remove the first message in the phrase */
	public MidiMsg removeFirstMsg() {
		if ( _msg == null )
			return null;
		MidiMsg m = _msg;
		if ( _lastmsg == _msg ) {
			_msg = null;
			_lastmsg = null;
		}
		else
			_msg = _msg.next;
		m.next = null;
		return m;	// we don't have to clone it, because we've pulled it off.
	}

	/** Remove a particular message from the phrase */
	public MidiPhrase removeMsg(MidiMsg mtoremove) {
		if ( mtoremove == null )
			return this;
		MidiMsg m;
		MidiMsg pre=null;
		for ( m=_msg; m!=null; pre=m,m=m.next ) {
			if ( m == mtoremove ) {
				if ( pre==null ) {
					 // It's the first msg

					// If it's also the last msg...
					if ( _lastmsg == _msg ) {
						_msg = null;
						_lastmsg = null;
					}
					else
						_msg = _msg.next;
				}
				else {
					pre.next = m.next;
				}
				break;
			}
		}
		return this;
	}

	/** Return the Nth (1-based) message of the phrase
	 * This method is 1-based to make it easier to convert Keynote code.
	 * Use getAt() for the 0-based version.
	 * @param n Number (1-based) of message to get
	 */
	public MidiMsg getNth(int n) {
		if ( n <= 0 )
			return null;
		return getAt(n-1);	// getAt will clone it
	}

	/** Return the Nth (0-based) message of the phrase
	 * @param n Index (0-based) of message to get
	 */
	public MidiMsg getAt(int n) {
		MidiMsg m = _msg;
		for ( int i=0; i<n && m != null; i++ )
			m = m.next;
		return (MidiMsg)m.clone();
	}

	/** Replace a particular message in the phrase with a new message.
	 * @param oldmsg MidiMsg to look for.
	 * @param newmsg MidiMsg to replace it with.
	 */
	public MidiPhrase replace(MidiMsg oldmsg, MidiMsg newmsg) {
		MidiMsg m;
		MidiMsg pre = null;
		for ( m=_msg; m != null; pre=m,m=m.next ) {
			if ( m == oldmsg ) {
				if ( newmsg == null ) {
					// delete oldmsg from list entirely
					if ( pre == null )
						_msg = m.next;
					else
						pre.next = m.next;
				}
				else {
					// replace oldmsg with newmsg
					newmsg.next = oldmsg.next;
					if ( pre == null )
						_msg = newmsg;
					else
						pre.next = newmsg;
				}
				break;
			}
		}
		_lastmsg = lastMsg();
		return this;
	}

	/** Return last msg in Phrase
	 */
	private MidiMsg lastMsg() {
		return lastMsg(_msg);
	}
	private MidiMsg lastMsg(MidiMsg m) {
		if ( m == null )
			return m;
		while ( m.next != null )
			m = m.next;
		return m;
	}


	/** Append a MidiMsg to the phrase
	 * The starting time of the added MidiMsg is set to the current
	 * length of this phrase, and the length of this phrase is
	 * incremented by the duration of the MidiMsg.
	 * @param n MidiMsg to append
	 */
	public MidiPhrase append(MidiMsg newm) {

		if ( newm == null )
			return this;
		newm = (MidiMsg)newm.clone();

		// Use _phraseLength of Phrase to determine time placement
		newm.setTime(newm.getTime()+_phraseLength);

		_phraseLength += newm.getDuration();

		MidiMsg m = _msg;
		if ( m == null ) {
			// Special case - it was an empty phrase to begin with
			_msg = newm;
		}
		else if ( _lastmsg != null ) {
			// Append to last msg
			_lastmsg.next = newm;
		}
		else {
			// Find last entry and append
			while ( m.next != null )
				m = m.next;
			m.next = newm;
		}
		_lastmsg = newm;
		return this;
	}

	/** Append a MidiPhrase to this phrase
	 * The starting time of the added MidiPhrase (i.e. of all the MidiMsgs
	 * in the phrase) is set to the current length of this phrase,
	 * and the length of this phrase is incremented by the length
	 * of the added phrase.
	 * @param newp MidiPhrase to append
	 */
	public MidiPhrase append(MidiPhrase newp) {

		newp = (MidiPhrase)newp.clone();

		// Use _phraseLength of Phrase to determine time placement
		newp.addTime(_phraseLength);

		_phraseLength += newp.getLength();

		// Special case of empty phrase
		MidiMsg m = _msg;
		if ( m == null ) {
			_msg = newp._msg;
		}
		else {
			// Find last entry and append
			while ( m.next != null )
				m = m.next;
			m.next = newp._msg;
		}
		_lastmsg = lastMsg();
		return this;
	}

	/** Merge a MidiMsg into this phrase
	 * The new MidiMsg is merged with this phrase, putting it in
	 * the appropriate time-ordered location within this phrase's
	 * list of MidiMsgs.
	 * @param newmsg MidiMsg to merge.
	 */
	public MidiPhrase merge(MidiMsg newmsg) {

		if ( newmsg==null )
			return this;

		newmsg = (MidiMsg)newmsg.clone();

		// special case for empty phrase
		if ( _msg == null ) {
			_msg = newmsg;
			_lastmsg = newmsg;
			return this;
		}

		MidiMsg m = _msg;
		MidiMsg pre = null;
		for ( ; ; pre=m, m=m.next ) {
			if ( m == newmsg ) {
				System.err.println("Hey, trying to merge a MidiMsg with a phrase that already contains it!?");
				break;
			}
			if ( newmsg.compareOrder(m) < 0 ) {
				if ( pre == null ) {
					// add to beginning of phrase
					newmsg.next = m;
					_msg = newmsg;
				}
				else {
					newmsg.next = m;
					pre.next = newmsg;
				}
				break;
			}
			if ( m.next == null ) {
				// We reached the end of the list, just append it.
				m.next = newmsg;
				long e = newmsg.time + newmsg.getDuration();
				if ( e > _phraseLength )
					_phraseLength = e;
				break;
			}
		}
		_lastmsg = lastMsg();
		return this;
	}

	/** Merge a Midiphrase into this phrase
	 * @param newphr MidiPhrase to merge.
	 */
	public MidiPhrase merge(MidiPhrase p) {

		if ( p == null || p._msg == null )
			return this;
		p = (MidiPhrase)p.clone();

		if ( p._phraseLength > _phraseLength )
			_phraseLength = p._phraseLength;

		if ( this._msg == null ) {
			this._msg = p._msg;
			this._lastmsg = p._lastmsg;
			return this;
		}

		// If the notes in p are all after the last note in this,
		// then we can do it in one fell swoop.
		if ( this._lastmsg.compareOrder(p._msg) < 0 ) {
			this._lastmsg.next = p._msg;
			this._lastmsg = p._lastmsg;
			return this;
		}
		MidiMsg m1 = this._msg;
		MidiMsg m2 = p._msg;

		// Get the first note into the result
		if ( m1.compareOrder(m2) < 0 ) {
			this._msg = m1;
			this._lastmsg = m1;
			m1 = m1.next;
			this._lastmsg.next = null;
		}
		else {
			this._msg = m2;
			this._lastmsg = m2;
			m2 = m2.next;
			this._lastmsg.next = null;
		}
		for ( ;; ) {
			// Put the lesser of the two available msgs onto
			// the result.
			if ( m1 == null ) {
				// Take the rest of m2, we're done.
				this._lastmsg.next = m2;
				this._lastmsg = lastMsg(m2);
				break;
			}
			else if ( m2 == null ) {
				// Take the rest of m1, we're done.
				this._lastmsg.next = m1;
				this._lastmsg = lastMsg(m1);
				break;
			}
			else if ( m1.compareOrder(m2) < 0 ) {
				this._lastmsg.next = m1;
				this._lastmsg = m1;
				m1 = m1.next;
				this._lastmsg.next = null;
			}
			else {
				this._lastmsg.next = m2;
				this._lastmsg = m2;
				m2 = m2.next;
				this._lastmsg.next = null;
			}
		}
		return this;
	}

	/** Return a copy of this phrase */
	public Object clone() {
		return cloneStartingFrom(_msg);
	}

	/** Return a copy of this phrase, starting at m */
	public Object cloneStartingFrom(MidiMsg m) {

		MidiPhrase newp = new MidiPhrase();
		newp._phraseLength = _phraseLength;
		if ( m == null )	// special case - empty phrase
			return newp;

		MidiMsg lastm = null;
		for ( MidiMsg oldm=m; oldm!=null; oldm=oldm.next ) {
			MidiMsg newm = (MidiMsg)oldm.clone();
			newm.next = null;
			if ( lastm == null )
				newp._msg = newm;
			else
				lastm.next = newm;
			lastm = newm;
		}
		newp._lastmsg = lastm;
		return newp;
	}

	public boolean equals(MidiPhrase p2) {
		Enumeration e1 = this.getEnumeration();
		Enumeration e2 = p2.getEnumeration();
		while ( e1.hasMoreElements() && e2.hasMoreElements() ) {
			MidiMsg m1 = (MidiMsg) e1.nextElement();
			MidiMsg m2 = (MidiMsg) e2.nextElement();
			if ( ! m1.equals(m2) )
				return false;
		}
		// does either one have more?  Then they're not equal
		if ( e1.hasMoreElements() || e2.hasMoreElements() )
			return false;
		else {
			return true;
		}
	}

	/** Set time of all MidiMsgs in this phrase
	 * @param tm time (in clicks)
	 */
	public MidiPhrase setTime(long tm) {
		for ( MidiMsg m=_msg; m!=null; m=m.next )
			m.setTime(tm);
		return this;
	}

	/** Set pitch of all MidiMsgs in this phrase
	 * @param p pitch (0-127)
	 */
	public MidiPhrase setPitch(int p) {
		for ( MidiMsg m=_msg; m!=null; m=m.next )
			m.setPitch(p);
		return this;
	}

	/** Set channel of all MidiMsgs in this phrase
	 * @param c channel (1-16)
	 */
	public MidiPhrase setChannel(int c) {
		for ( MidiMsg m=_msg; m!=null; m=m.next )
			m.setChannel(c);
		return this;
	}

	/** Set duration of all MidiMsgs in this phrase
	 * @param d duration (in clicks)
	 */
	public MidiPhrase setDuration(long d) {
		for ( MidiMsg m=_msg; m!=null; m=m.next )
			m.setDuration(d);
		return this;
	}

	/** Add a time to all MidiMsgs in this phrase
	 * @param tm time (in clicks)
	 */
	public MidiPhrase addTime(long tm) {
		for ( MidiMsg m=_msg; m!=null; m=m.next )
			m.time += tm;
		return this;
	}

	/** Adjust by a factor the velocity of all MidiMsgs in this phrase
	 * @param f multiplication factor
	 */
	public MidiPhrase multiplyVelocity(double f) {
		for ( MidiMsg m=_msg; m!=null; m=m.next ) {
			int v = (int)(m.getVelocity()*f);
			if ( v < 0 )
				v = 0;
			if ( v > 127 )
				v = 127;
			m.setVelocity(v);
		}
		return this;
	}

	/** Make sure the order of MidiMsgs in this phrase is correct.
	 * This is used after doing appending and other things that may
	 * have resulted in things being out of order.
	 */
	public MidiPhrase reOrder() {
		if ( _msg == null || _msg.next == null )
			return this;

		MidiMsg pre = _msg;
		MidiMsg m = pre.next;
		for ( ; m!=null && m.next!=null; ) {

			// If next note is out of order...
			if ( m.compareOrder(pre) < 0 ) {

				// Remove m(which we've discovered should be before pre)
				pre.next = m.next;
				this.merge(m);
				m = pre.next;
			}
			else {
				pre = m;
				m = m.next;
			}
		}
		_lastmsg = lastMsg();
		return this;
	}

	/** return canonical octave */
	private int canoctave(int p) {
		return -2 + (p/12);
	}

	/** return canonical pitch */
	private int canonipitch(int p) {
		return p%12;
	}

	static String Scachars[];
	static {
		Scachars = new String[13];
		Scachars[0] = "c";
		Scachars[1] = "c+";
		Scachars[2] = "d";
		Scachars[3] = "e-";
		Scachars[4] = "e";
		Scachars[5] = "f";
		Scachars[6] = "f+";
		Scachars[7] = "g";
		Scachars[8] = "a-";
		Scachars[9] = "a";
		Scachars[10] = "b-";
		Scachars[11] = "b";
		Scachars[12] = "c";
	}

	/** Return a human-readable version of this phrase, in KeyKit format. */
	public String toString() {
		MidiMsg m = _msg;
		String s = "'";
		resetdef();
		boolean first = true;
		long maxend = 0;
		while ( m != null ) {
			if ( first )
				first = false;
			else {
				String sep;
				if ( m.getTime() == Deftime )
					sep = " ";
				else {
					sep = ",";
					Deftime = Def2time;
				}
				s += sep;
			}
			if ( ! (m instanceof MidiNoteable) ) {
				s += m.toString();
			}
			else {
				if ( m instanceof MidiNoteOn )
					s += "+";
				else if ( m instanceof MidiNoteOff )
					s += "-";

				s += Scachars[canonipitch(m.getPitch())];

				int oct = canoctave(m.getPitch());
				if ( oct != Defoct )
					s += ("o" + (Defoct=oct));

				long dur = m.getDuration();
				if ( dur != Defdur )
					s += ("d" + (Defdur=dur));

				int vol = m.getVelocity();
				if ( vol != Defvol )
					s += ("v" + (Defvol=vol));

				int chan = m.getChannel();
				if ( chan != Defchan )
					s += ("c" + (Defchan=chan));
			}
			if ( m.getTime() != Deftime ) {
				s += "t"+m.getTime();
			}
			Deftime = m.getTime();
			Def2time = Deftime + m.getDuration();
			if ( Def2time > maxend )
				maxend = Def2time;
			m = m.next;
		}
		s += "'";
		return s;
	}

	void resetdef() {
		// This determines the default values at the beginning of a 
		// constant phrase string.
		Defvol = 64;
		Defoct = 3;
		Defdur = 96;
		Defchan = 1;	// channel is 1-based
		Deftime = Def2time = 0;
	}

	/** Interpret a KeyKit-style string and merge it. */
	void mergeKeyKitValue(String s) {

		resetdef();

		StringTokenizer toke = new StringTokenizer(s,", '",true);
		while ( toke.hasMoreTokens() ) {
String ttt;
			ScanState state = new ScanState(ttt=toke.nextToken());
			int c = state.peekChar();
			switch (c) {
			case ',':
				// a comma means the next note starts where the previous ended
				Deftime = Def2time;
				break;
			case ' ':
				break;
			case '\'':	// Single quotes should only be at beginning and end,
						// if anywhere.
				break;
			default:
				MidiMsg n = noteParse(state);
				if ( n != null )
					merge(n);
				break;
			}
		}
	}

	class ScanState {
		String str;
		int pos;
		int value;
		int leng;
		ScanState(String s) {
			str = s;
			pos = 0;
			leng = str.length();
			value = 0;
		}
		int peekChar() {
			if ( pos >= leng )
				return -1;
			else
				return str.charAt(pos);
		}
		void skipChar() {
			if ( pos < leng )
				pos++;
		}
		int nextChar() {
			if ( pos >= leng )
				return -1;
			else
				return str.charAt(pos++);
		}
		int numscan() {
			int v = 0;
			for ( int nscanned=0;; nscanned++ ) {
				int c = peekChar();
				if ( c < 0 )
					break;
				int i = "0123456789".indexOf(c);
				if ( i < 0 )
					break;
				skipChar();	// we've already got it
				v = v * 10 + i;
			}
			return v;
		}
	}

	private MidiMsg noteParse(ScanState state) {

		MidiMsg n;

		int c = state.peekChar();
		switch (c) {

		case 'l':
			state.skipChar();
			// explicit phrase length
			_phraseLength = state.numscan();
			return null;

		case 'x':
			state.skipChar();
			// raw midi bytes
			n  = stringToMidiBytes(state);
			break;

		case 'r':
			// rest - it gets scanned, to update Deftime
			stringToNote(' ',state);
			// but its of no other use
			return null;

		case '+':
		case '-':
			state.skipChar();
			n = stringToNote(c,state);
			break;

		default:
			n = stringToNote(' ',state);
			break;
		}

		Deftime = n.getTime();
		Def2time = n.getTime() + n.getDuration();

		return n;
	}

	private int hexchar(int c) {
		int v = "0123456789abcdef".indexOf(c);
		if ( v < 0 ) {
			v = "0123456789ABCDEF".indexOf(c);
			if ( v < 0 )
				v = 0;
		}
		return v;
	}

	private MidiBytes stringToMidiBytes(ScanState state) {
		MidiBytes n = new MidiBytes();
		n.setTime(Deftime);
		long tm;
		while (true) {
			int c1 = state.nextChar();
			if ( c1 < 0 )
				break;
			if ( c1 == 't' ) {
				n.setTime(state.numscan());
				// it's supposed to always be the last thing
				break;
			}
			// it's the first of a hex pair
			int c2 = state.nextChar();
			if ( c2 < 0 )
				break;
			int v1 = hexchar(c1);
			int v2 = hexchar(c2);
			if ( v1 >= 0 && v2 >= 0 )
				n.addByte((byte)(v1 * 16 + v2));
		}
		return n;
	}

	static int[] pitchvals;
	static {
		// This array matches the pitch values of
		// notes 'c', 'd', 'e', through 'b' (octave 0)
		pitchvals = new int[7];
		pitchvals[0] = 24;
		pitchvals[1] = 26;
		pitchvals[2] = 28;
		pitchvals[3] = 29;
		pitchvals[4] = 31;
		pitchvals[5] = 33;
		pitchvals[6] = 35;
	}

	private MidiMsg stringToNote(int type, ScanState state) {

		boolean isrest = false;
		int val = 0;
		int vol = Defvol;
		int octave = Defoct;
		long dur = Defdur;
		int chan = Defchan;
		long clicks = Deftime;
		int pitch = -1;

		int c = state.nextChar();
		if ( c == 'p' ) {
			// it's a raw pitch value
			pitch = state.numscan();
		}
		else if ( c == 'r' ) {
			// It's a rest
			isrest = true;
		}
		else {
			// it's a pitch name
			int si = "cdefgab".indexOf(c);
			if ( si < 0 )
				return null;
			val = pitchvals[si];
		}
		while ( (c=state.nextChar()) >= 0 ) {
			switch (c) {
			case '+':
			case '#':
				val++;
				break;
			case '-':
				val--;
				break;
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				octave = c - '0';
				break;
			case 'o':
			case 'O':
				octave = state.numscan();
				break;
			case 'v':
			case 'V':
				vol = state.numscan();
				break;
			case 'd':
			case 'D':
				dur = state.numscan();
				break;
			case 't':
			case 'T':
				clicks = state.numscan();
				break;
			case 'c':
			case 'C':
				chan = state.numscan();
				break;
			}
		}
		/* If a 'raw' pitch number wasn't given .. */
		if ( pitch < 0 )
			pitch = val + 12 * octave;

		/* various sanity checking */
		if ( pitch < 0 )
			pitch = 0;
		else if ( pitch > 127 )
			pitch = 127;

		if ( chan < 1 )
			chan = 1;
		else if ( chan > 16 )
			chan = 16;

		if ( vol < 0 )
			vol = 0;
		else if ( vol > 127 )
			vol = 127;

		if ( dur < 0 )
			dur = 0;

		Defchan = chan;
		Defdur = dur;
		Defvol = vol;
		Defoct = octave;

		/* For rests, we just update Deftime* so the default for the */
		/* next note will be appropriate. */
		if ( isrest ) {
			Deftime = Def2time = clicks + dur;
			return null;
		}

		MidiMsg n;
		switch(type) {
		case '+':
			n = new MidiNoteOn(pitch,clicks,chan,vol);
			break;
		case '-':
			n = new MidiNoteOff(pitch,clicks,chan,vol);
			break;
		default:
			n = new MidiNote(pitch,clicks,chan,vol,dur);
			break;
		}
		return n;
	}

	public static void main(String[] args) {
		MidiPhrase p = new MidiPhrase(args[0]);
		System.out.println("p="+p);
	}
}
