/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * MidiSystem is the main class for controlling the Midi package.
 * All definitions and methods are static.
 * To use MIDI, invoke:
 * <pre><center>
 * MidiSystem.start();
 * </center></pre>
 * and then you can schedule MIDI output by doing things like:
 * <pre><center>
 * MidiSystem.schedule(new MidiNote(100));   // pitch is 100
 * </center></pre>
 */

public class MidiSystem {

	static private MidiOutput outDevice;
	static private MidiInput inDevice;
	static private long time0;
	static private boolean started = false;
	static private int microPerBeat = 500000;
	static private int clksPerBeat = 96;
	static private boolean interfaceLoaded = false;
	static private MidiAction last_act;
	static private int last_inId;
	static private int last_outId;
	static private MidiSystem singleton;
	static private String dllname = "nosuchmidi";

	static {
		/* try to load the DLL containing the MIDI interface */
		try {
			System.loadLibrary(dllname);
			interfaceLoaded = true;
		} catch (UnsatisfiedLinkError e) {
			System.err.println(dllname+" library not found");
		}
	}

	static private void initsingleton() {
		// The main purpose of this is for native methods, since
		// I couldn't get static native methods to work.
		if ( singleton == null )
			singleton = new MidiSystem();
	}

	/** return the number of input devices */	
	static public int getNumInDevices() {
		initsingleton();
		try {
		   	return singleton.nativeMidiInGetNumDevs();
		} catch (Exception e) {
			return -1;
		}
	}

	/** return the number of output devices */	
	static public int getNumOutDevices() {
		initsingleton();
		try {
		   	return singleton.nativeMidiOutGetNumDevs();
		} catch (Exception e) {
			return -1;
		}
	}

	/** Return true if MidiSystem has been started. */
	public static boolean isStarted() {
		return started;
	}

	/** Initialize the MIDI system with default input and output devices. */
	public static boolean start() {
		return start(null,-1,0);
	}

	/** Initialize the MIDI system with specific callback actions.
	 * @param act MidiAction for callbacks on MIDI input messages.
	 */
	public static boolean start(MidiAction act) {
		return start(act,-1,0);
	}

	/** Initialize the MIDI system with specific callbacks and I/O devices.
	 * @param act MidiAction for callbacks on MIDI input messages.
	 * @param outId Id of MIDI output device
	 * @param inId Id of MIDI input device
	 */
	public static boolean start(MidiAction act, int outId, int inId) {

		if ( interfaceLoaded == false ) {
			System.err.println(dllname+" library not loaded");
			outDevice = null;
			return false;
		}

		// Applets don't necessarily have to call MidiSystem.stop when
		// they stop, so when an applet starts, the MIDI system may already
		// be in good enough shape (with the proper input/output devices
		// and actions set up) to re-use.  The safe approach would just
		// be to automatically call stop() if started==true,
		// however this causes a delay and (more importantly) a loud POP
		// on my system.  With the approach implemented below, if you
		// re-visit a MIDI applet page that has the same needs as a previous
		// MIDI applet page, it will just re-use the running system.

		if ( started==true && act == last_act
			&& last_outId == outId && last_inId == inId ) {

				// Already started, and since attributes match the
				// last use, we don't have to stop/re-start
				return true;
		}

		if ( started == true ) {
			stop();
		}

		last_act = act;
		last_outId = outId;
		last_inId = inId;

		time0 = System.currentTimeMillis();

		try {
			outDevice = new Win32MidiOutput(outId);
			outDevice.open();
			outDevice.startRunning();
		} catch (MidiException e) {
			System.err.println("Error opening MidiOutput - "+e.getMessage());
			outDevice = null;
		}

		try {
			inDevice = new Win32MidiInput(inId);
			inDevice.open();
			inDevice.setAction(act);
			inDevice.startRunning();
		} catch (MidiException e) {
			System.err.println("Error opening MidiInput - "+e.getMessage());
			inDevice = null;
		}

		// IS THIS STILL NEEDED?
		if ( outDevice != null || inDevice != null ) {
			// Sleep for a tiny bit to let the threads get going.
			// This could be replaced with a little more direct approach.
			// Basically it's to allow the threads to get to the point
			// where they're blocked waiting for events.
			try {
				Thread.currentThread().sleep(10);
			} catch (Exception e) {
				System.err.println("error while sleeping? - "+e.getMessage());
			}
			started = true;
		}
		return started;
	}

	/** Set the output device to a new device (this is untested) */
	public static void setOutDevice(MidiOutput o) {
		try {
			if ( outDevice != null )
					outDevice.close();
			outDevice = o;
			outDevice.open();
		} catch (MidiException e) {
			System.err.println("Error setting OutDevice - "+e.getMessage());
			outDevice = null;
		}
	}

	/** Set the input device to a new device (this is untested) */
	public static void setInDevice(MidiInput o) {
		try {
			if ( inDevice != null )
					inDevice.close();
			inDevice = o;
			inDevice.open();
		} catch (MidiException e) {
			System.err.println("Error setting InDevice - "+e.getMessage());
			inDevice = null;
		}
	}

	/**
	 * Stop the MIDI system.
	 */
	public static void stop() {
		if ( outDevice != null ) {
			try {
				outDevice.stopRunning();
				outDevice.close();
				outDevice = null;
            } catch (MidiException e) {
                System.err.println("Closing MIDI output - "+e.getMessage());
            }
		}
		if ( inDevice != null ) {
			try {
				inDevice.stopRunning();
				inDevice.close();
				inDevice = null;
            } catch (MidiException e) {
                System.err.println("Closing MIDI input - "+e.getMessage());
            }
		}
		started = false;
		last_act = null;
		last_outId = -99;
		last_inId = -99;
	}

	/** Return a phrase containing allo-notes-off messsages for all channels
	 * Currently this is only on the default channel */
	public static MidiPhrase allNotesOff() {
		MidiPhrase p = new MidiPhrase();
		for ( int i=0; i<16; i++ ) {
			p.append(new MidiBytes((byte)(i|0xb0),(byte)(0x7b),(byte)0));
		}
		return p;
	}

	/**
	 * Return number of milliseconds in a specified number of clicks.
	 * @param nclicks number of clicks
	 */
	public static long clickToMilli(long nclicks) {
		return ((nclicks * microPerBeat) / clksPerBeat) / 1000;
	}

	/**
	 * Return number of milliseconds in a specified number of beats.
	 * @param nbeats number of beats
	 */
	public static long beatToMilli(long nbeats) {
		return (nbeats * microPerBeat) / 1000;
	}

	/**
	 * Return number of milliseconds in one beat.
	 */
	public static long milliPerBeat() {
		return (microPerBeat) / 1000;
	}

	/**
	 * Return number of clicks in a specified number of milliseconds.
	 * @param milli number of milliseconds
	 */
	public static long milliToClick(long milli) {
		return (milli * 1000) * clksPerBeat / microPerBeat;
	}

	/**
	 * Return number of clicks in a beat.
	 */
	public static int clicksPerBeat() {
		return clksPerBeat;
	}

	/**
	 * Return number of clicks in a specified number of beats.
	 * @param nbeats number of beats
	 */
	public static long beatToClick(long nbeats) {
		return nbeats * clksPerBeat;
	}

	/**
	 * Schedule a MidiMsg for eventual output.
	 * @param m MidiMsg to be scheduled
	 * @exception MidiException If unable to schedule
	 */
	public synchronized static void schedule(MidiMsg m) throws MidiException {
		schedule(new MidiPhrase(m));
	}

	/**
	 * Schedule a MidiPhrase for eventual output
	 * The time offset of the phrase is set to MidiSystem.now(), so that
	 * if the time of the phrase starts at 0 (which is pretty normal),
	 * the phrase will play immediately.
	 * @param p MidiPhrase to be scheduled
	 * @exception MidiException If unable to schedule
	 */
	public synchronized static void schedule(MidiPhrase p) throws MidiException {
		schedule(p,now());
	}

	/**
	 * Schedule a MidiPhrase for eventual output.
	 * The offset value controls when the phrase will begin.  If the
	 * phrase time (i.e. the times attached to its MidiMsg objects)
	 * starts at 0, and the offset value is set to MidiSystem.now(),
	 * then playback will start immediately.
	 * @param p MidiPhrase to be scheduled
	 * @param offset An offset (in clicks) to be added to the time of the phrase
	 * @exception MidiException If unable to schedule
	 */
	public synchronized static void schedule(MidiPhrase p, long offset) throws MidiException{
		if ( started == false )
			throw new MidiException("MidiSystem not started");
		if ( outDevice == null )
			throw new MidiException("Can't schedule - No outDevice");
		outDevice.schedule(p,offset);
	}

	public synchronized static void schedule(byte[] b) throws MidiException{
		schedule(b,now());
	}

	public synchronized static void schedule(byte[] b, long offset) throws MidiException{
		if ( started == false )
			throw new MidiException("MidiSystem not started");
		if ( outDevice == null )
			throw new MidiException("Can't schedule - No outDevice");
		outDevice.schedule(b,offset);
	}

	/**
	 * Set the tempo of schedule playback.
	 * @param	t	tempo in microseconds per beat.
	 * @exception MidiException If something goes wrong.
	 */
	public static void setTempo(int t) throws MidiException {
		initsingleton();
		microPerBeat = t;
		singleton.nativeSetTempo(t);
	}

	/**
	 * Set the number of clicks per beat.
	 * @param	c	Number of clicks per beat.
	 * @exception MidiException If something goes wrong.
	 */
	public static void setClicks(int c) throws MidiException {
		initsingleton();
		singleton.nativeSetClicks(c);
	}

	/**
	 * Set the debug level in the native methods.
	 * @param	d	Debug level.
	 * @exception MidiException If something goes wrong.
	 */
	public static void setDebug(int d) throws MidiException {
		initsingleton();
		singleton.nativeSetDebug(d);
	}
	
	/**
	 * Return the current time in clicks.
	 * @returns	Current time in clicks.
	 */
	public static long now() {
		initsingleton();
		return singleton.nativeNow();
	}
	
	/**
	 * Native Methods
	 */

	/**
	 * set the internal Tempo of the scheduler
	 * @exception MidiException If something goes wrong.
	 */
	public native void nativeSetTempo(int t) throws MidiException;

	/**
	 * set the value of Clicks for the scheduler
	 * @exception MidiException If something goes wrong.
	 */
	public native void nativeSetClicks(int c) throws MidiException;

	/**
	 * Set the debug flag for the Midi system
	 */
	public native void nativeSetDebug(int d);

	/**
	 * Return value of current time in clicks (not milliseconds).
	 * @returns	Current time in clicks.
	 */
	public native long nativeNow();

	/**
	 * Query the system regarding the number of MIDI input devices
	 * @returns The number of MIDI input devices.
	 * @exception MidiException If something goes wrong.
	 */
	public native int nativeMidiInGetNumDevs() throws MidiException;

	/**
	 * Query the system regarding the number of MIDI output devices
	 * @exception MidiException If something goes wrong.
	 * @returns The number of MIDI output devices.
	 */
	public native int nativeMidiOutGetNumDevs() throws MidiException;

	/**
	 * open a URL in a browser
	 * @exception MidiException If something goes wrong.
	 */
	public native void nativeOpenURL(String browser, String url) throws MidiException;

	public static void main(String[] args) {
		// System.out.println("MidiSystem main() start");
		boolean ok = MidiSystem.start();
		if ( ! ok ) {
			System.err.println("Can't initialize MIDI !?");
			System.exit(1);
		}
		MidiPhrase p = new MidiPhrase("a,b,c");
		try {
			MidiSystem.schedule(new MidiNote(80));
			MidiSystem.schedule(p);
		} catch (MidiException e) {
			System.err.println("MidiException e="+e.getMessage());
		}
		// System.out.println("MidiSystem main() end");
	}
}

