/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * Handle MIDI output.
 */

public class Win32MidiOutput implements MidiOutput, Runnable {

	int deviceId;
	int deviceHandle = 0;
	Thread t;
	boolean realtime;

	public static final int MIDIMAPPER = -1;

	/**
	 * Contructor for using MIDIMAPPER device.
	 * @exception MidiException If anything goes wrong.
	 */
	public Win32MidiOutput() throws MidiException {
		this(MIDIMAPPER);
	}

	/**
	 * Contructor for using a particular device
	 * @param id	The device id.
	 * @exception MidiException If anything goes wrong.
	 */
	public Win32MidiOutput(int id) throws MidiException {
		this(id,true);
	}

	/**
	 * Contructor for using a particular device
	 * @param id	The device id.
	 * @param r	True if you want to enable windows realtime priority.
	 * @exception MidiException If anything goes wrong.
	 */
	public Win32MidiOutput(int id, boolean r) throws MidiException {
		deviceId = id;
		realtime = r;

		// MIDIMAPPER is a Windows-defined MIDI filter used for performing MIDI
		// message translation on the fly.
		if (id < 0 && id != MIDIMAPPER) {
			throw new MidiException("Bad output deviceId - "+Integer.toString(id));
		}
	}

	/**
	 * Open the device.
	 * @exception MidiException If anything goes wrong.
	 */
	public void open() throws MidiException {

		// only try to open if it is not already open	
 		if (deviceHandle != 0)
			return;

		// attempt to open the device
		MidiOutOpen();
	}


	/**
	 * Reset the MIDI out device.  This would also be the logical place to turn
	 * notes off if we add note tracking.
	 */
	public void reset() {
		try {
		   	MidiOutReset();
		} catch (Exception e) {
			System.err.println("Error in MidiOutReset - "+e.getMessage());
		}
	}

	/**
	 * Close the device.
	 * @exception MidiException If anything goes wrong.
	 */
	public synchronized void close() throws MidiException {

		if (deviceHandle == 0) {	// not open; nothing to do
			return;
		}
	
		// reset and close midi out
		reset();
		MidiOutClose();
	}

	public void startRunning() {
		t = new Thread(this);
		t.start();
	}

	/**
	 * Run the output loop.
	 */
	public void run() {
		Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
		try {
			MidiOutLoop(realtime);
		} catch (Exception e) {
			System.err.println("MidiOutLoop error - "+e.getMessage());
		}
	}

	/**
	 * Stop the output loop.
	 */
	public void stopRunning() {
		try {
			MidiOutLoopStop();
		} catch (Exception e) {
			System.err.println("MidiOutLoopStop error!? - "+e.getMessage());
		}
		t.stop();
	}

	/**
	 * Schedule a MidiPhrase for output.
	 * @param p MidiPhrase to schedule
	 * @param offset Time (in clicks) to add to phrase before scheduling
	 * @exception MidiException If anything goes wrong.
	 */
	public synchronized void schedule(MidiPhrase p,long offset) throws MidiException{

		// The time values in the phrases we're being given are
		// incremented by "offset", and are then considered as being
		// relative to time0.  The time values are in Clicks,
		// NOT MILLISECONDS.

		MidiPhraseEnumeration e = new MidiPhraseEnumeration(p);
		while ( e.hasMoreElements() ) {
			MidiMsg m = (MidiMsg)e.nextElement();
			byte[] b = m.getBytesOn();
			if ( b != null ) {
				scheduleBytes((long)(m.time+offset),b);
				b = m.getBytesOff();
				if ( b != null ) {
					scheduleBytes((long)(m.time+offset+m.getDuration()),b);
				}
			}
		}
	}

	/**
	 * Schedule bytes for output.
	 * @param bytes Bytes to schedule
	 * @param offset Time (in clicks) to add to phrase before scheduling
	 * @exception MidiException If anything goes wrong.
	 */
	public synchronized void schedule(byte[] b,long offset) throws MidiException {
		scheduleBytes(offset,b);
	}


	/**
	 * Native Methods
	 */
	
	/**
	 * @exception MidiException If anything goes wrong.
	 */
	private native void MidiOutOpen() throws MidiException;

	/**
	 * @exception MidiException If anything goes wrong.
	 */
	private native int MidiOutReset() throws MidiException;

	/**
	 * @exception MidiException If anything goes wrong.
	 */
	private native int MidiOutClose() throws MidiException;

	/**
	 * @exception MidiException If anything goes wrong.
	 */
	private native void MidiOutLoop(boolean r) throws MidiException;

	private native void MidiOutLoopStop() throws MidiException;

	/**
	 * @exception MidiException If anything goes wrong.
	 */
	private native void scheduleBytes(long tm, byte[] b) throws MidiException;

}
